//*****************************************************************************
// File: NvClassLoader.java
// Description: Nirva Java ClassLoader
// Author: Lionel Martin
// Copyright (C) 2013 Nirva Software
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

package com.nirvasoft.nirva;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.AccessControlContext;
import java.security.AccessController;
import java.security.CodeSource;
import java.security.PrivilegedAction;
import java.security.PrivilegedExceptionAction;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Locale;
import java.util.jar.Attributes;
import java.util.jar.Attributes.Name;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

public class NvClassLoader extends URLClassLoader {
	
	private AccessControlContext acc;
	
	private File[] folders = new File[0];
	private File[] jarFiles = new File[0];
	private JarFile[] jars = new JarFile[0];
	
	private HashMap<String, ResourceCacheEntry> resources = new HashMap<String, ResourceCacheEntry>();

	public NvClassLoader(URL[] urls) {
		super(new URL[0]);
		try {
			for (URL url : urls) {
//				System.out.println("Found URL: "+url.toString()+", Protocol '"+url.getProtocol()+"', File ("+url.getFile()+")");
				addURL(url);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public NvClassLoader(URL[] urls, ClassLoader parent) {
		super(new URL[0], parent);
		try {
			for (URL url : urls) {
//				System.out.println("Found URL: "+url.toString()+", Protocol '"+url.getProtocol()+"', File ("+url.getFile()+")");
				addURL(url);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

//	public NvClassLoader(URL[] urls, ClassLoader parent) {
//		super(urls, parent);
//		System.out.println("NvClassLoader with parent");
//		for (URL url : urls) {
//			System.out.println("Found URL: "+url.toString());
//		}
//	}
//
//	public NvClassLoader(URL[] urls, ClassLoader parent, URLStreamHandlerFactory factory) {
//		super(urls, parent, factory);
//		System.out.println("NvClassLoader with parent and factory");
//		for (URL url : urls) {
//			System.out.println("Found URL: "+url.toString());
//		}
//	}
	synchronized private void addRepository(String file, File folder) {
		if (file == null || folder == null) {
			return;
		}
		
		File[] newFolderList = new File[folders.length+1];
		for (int i = 0; i < folders.length; i++) {
			newFolderList[i] = folders[i];
		}
		try {
			folder = folder.getCanonicalFile();
		} catch (IOException e) {
			// Ignore
			e.printStackTrace(); // For debug
		}
		newFolderList[folders.length] = folder;
		folders = newFolderList;
	}
	synchronized private void addJar(String file, File jarFile) {
		if (file == null || jarFile == null) {
			return;
		}
		
		JarFile jar;
		try {
			jar = new JarFile(jarFile);
		} catch (IOException e) {
            // Catch the exception if there is an empty jar file
            // Should ignore and continue loading other jar files 
            // in the dir
			e.printStackTrace();
			return;
		}
		JarFile[] newJarList = new JarFile[jars.length+1];
		for (int i = 0; i < jars.length; i++) {
			newJarList[i] = jars[i];
		}
		newJarList[jars.length] = jar;
		jars = newJarList;
		
		File[] newJarFileList = new File[jarFiles.length+1];
		for (int i = 0; i < jarFiles.length; i++) {
			newJarFileList[i] = jarFiles[i];
		}
		try {
			jarFile = jarFile.getCanonicalFile();
		} catch (IOException e) {
			// Ignore
			e.printStackTrace(); // For debug
		}
		newJarFileList[jarFiles.length] = jarFile;
		jarFiles = newJarFileList;
		
	}

	@Override
	protected void addURL(URL url) {
//		try {
//			System.out.println("Adding a URL ("+url.toString()+") URI is ("+url.toURI().toString()+")");
//		} catch (URISyntaxException e) {
//			System.out.println("The URI stinks!!");
//			e.printStackTrace();
//		}
		if (!"file".equals(url.getProtocol())) {
			System.err.println("Protocol("+url.getProtocol()+") not handled. ["+url.toString()+"]");
			return;
		}
		String file = url.getFile().toLowerCase(Locale.ENGLISH);
		if (file.endsWith(".jar")) {
			File jar = new File(url.getFile());
			addJar(file, jar);
		} else if (file.endsWith("/") || file.endsWith("\\")) {
			File dir = new File(url.getFile());
			if (dir.isDirectory()) {
				addRepository(file, dir);
			} else {
				// TODO manage this error...
				System.err.println("'"+dir.getPath()+"' is not a directory ???");
			}
//			try {
//			} catch (URISyntaxException e) {
//	            // Catch the exception if there is an empty jar file
//	            // Should ignore and continue loading other jar files 
//	            // in the dir
//				e.printStackTrace();
//			}
		} else {
			// TODO manage this error...
			System.err.println("'"+file+"' is unknown...");
		}
//		super.addURL(url);
	}

	@Override
	public URL[] getURLs() {
		int nbFolders = folders.length;
		int nbJars = jars.length;
		
		try {
			URL[] urls = new URL[nbFolders+nbJars];
			for (int i = 0; i < nbFolders; i++) {
				File file = folders[i];
				urls[i] = file.toURI().toURL();
			}
			for (int i = 0; i < nbJars; i++) {
				File jarFile = jarFiles[i];
				urls[nbFolders+i] = jarFile.toURI().toURL();
			}
			return urls;
		} catch (MalformedURLException e) {
			return new URL[0];
		}
	}

	@Override
	protected Class<?> findClass(final String name) throws ClassNotFoundException {
//		System.out.println("findClass("+name+")");
		if (name == null) {
			return null;
		}
		
		try {
			return AccessController.doPrivileged(new PrivilegedExceptionAction<Class<?>>() {
				public Class<?> run() throws ClassNotFoundException {
					return findClass0(name);
				}
			}, acc);
		} catch (java.security.PrivilegedActionException pae) {
			throw (ClassNotFoundException) pae.getException();
		}
	}
	private Class<?> findClass0(String name) throws ClassNotFoundException {
		String path = name.replace('.', '/') + ".class";
		ResourceCacheEntry res = findResource0(name, path);
		if (res == null) {
			throw new ClassNotFoundException(name);
		}
		
		Class<?> clazz = res.clazz;
		if (clazz != null) {
			return clazz;
		}
		
		synchronized (this) {
			clazz = res.clazz;
			if (clazz != null) {
				return clazz;
			}
			
			if (res.binaryContent == null) {
				throw new ClassNotFoundException(name);
			}
			
			int lastDotIndex = name.lastIndexOf('.');
			String packageName = null;
			if (lastDotIndex != -1) {
				packageName = name.substring(0, lastDotIndex);
			}
			if (packageName != null) {
				Package pkg = getPackage(packageName);
				Manifest m = res.manifest;
				if (pkg == null) {
					if (m != null) {
						definePackage(packageName, m, res.getCodeBase());
					} else {
						definePackage(packageName, null, null, null, null, null, null, null);
					}
					pkg = getPackage(packageName);
				}
				
				if (pkg != null) { // Should be always ???
					boolean sealCheck = true;
					if (pkg.isSealed()) {
						sealCheck = pkg.isSealed(res.getCodeBase());
					} else {
						sealCheck = (m == null) || !isSealed(packageName, m);
					}
					if (!sealCheck) {
						throw new SecurityException("Sealing violation loading " + name + " : Package " + packageName + " is sealed.");
					}
				}
			}
			
            try {
            	clazz = defineClass(name, res.binaryContent, 0, res.binaryContent.length, new CodeSource(res.getCodeBase(), res.certificates));
            } catch (UnsupportedClassVersionError ucve) {
            	throw new UnsupportedClassVersionError(ucve.getLocalizedMessage() + " " + "NvClassLoader: unable to load class (" + name + ")");
            }
            
			res.clazz = clazz;
			res.binaryContent = null;
			res.manifest = null;
			res.certificates = null;
			res.setUrlsToNull();
		}
		return clazz;
	}
	/*
	 * Returns true if the specified package name is sealed according to the
	 * given manifest.
	 */
	private boolean isSealed(String name, Manifest man) {
		String path = name.replace('.', '/') + "/";
		Attributes attr = man.getAttributes(path);
		String sealed = null;
		if (attr != null) {
			sealed = attr.getValue(Name.SEALED);
		}
		if (sealed == null) {
			if ((attr = man.getMainAttributes()) != null) {
				sealed = attr.getValue(Name.SEALED);
			}
		}
		return "true".equalsIgnoreCase(sealed);
	}

	private ResourceCacheEntry findResource0(String name, String path) {
		ResourceCacheEntry entry = null;
		InputStream is = null;
		int entryContentLength = -1;
		boolean isClass = path.endsWith(".class");
		
		try {
			for (int i = 0; (entry == null) && (i < folders.length); i++) {
				String fullPath = folders[i].getAbsolutePath() + "/" + path;
//				System.out.println("Searching Class "+name+" at path "+fullPath);
				File file = new File(fullPath);
				try {
					file = file.getCanonicalFile();
				} catch (IOException e) {
					// Ignore
					e.printStackTrace(); // For debug...
				}
				if (file.exists()) {
					try {
						entry = new ResourceCacheEntry(file.toURI().toURL());
					} catch (MalformedURLException e) {
						e.printStackTrace(); // For debug...
						entry = null;
					}
					if (file.isFile()) {
						entryContentLength = (int) file.length();
						try {
							is = new FileInputStream(file);
						} catch (FileNotFoundException e) {
							// Ignore, should not happen...
							is = null;
							e.printStackTrace(); // For debug...
						}
//						entryContentLength = is.
					}
				}
//				if (entry == null) { // For debug
//					System.out.println("Searching Class "+name+" in folders failed...");
//				}
			}
			
			JarEntry jarEntry = null;
			synchronized (jars) {
				for (int i = 0; (entry == null) && (i < jars.length); i++) {
//					System.out.println("Searching Class "+name+" in jar "+jars[i].getName());
					jarEntry = jars[i].getJarEntry(path);
					if (jarEntry != null) {
						try {
							URL codeBase = jarFiles[i].toURI().toURL();
							String entryUrlInJar = "jar:" + codeBase.toString() + "!/" + path;
//							System.out.println("Searching Class "+name+" at path(jar) "+entryUrlInJar);
							entry = new ResourceCacheEntry(new URL(entryUrlInJar), codeBase);
						} catch (MalformedURLException e) {
							e.printStackTrace();// Just for debug
							entry = null;
						}
						entryContentLength = (int) jarEntry.getSize();
						try {
							entry.manifest = jars[i].getManifest();
							is = jars[i].getInputStream(jarEntry);
						} catch (IOException e) {
							e.printStackTrace();
							return null;
						}
					}
				}
				
				if (entry == null) {
//					System.out.println("Nothing found for "+name+" at '"+path+"'");
					return null;
				}
				
				if (is != null && isClass) {
					byte[] entryBinaryContent = new byte[entryContentLength];
					try {
						int pos = 0;
						while (true) {
							int nbRead = is.read(entryBinaryContent, pos, entryContentLength-pos);
							if (nbRead <= 0) {
								break;
							}
							pos += nbRead;
						}
					} catch (IOException e) {
						e.printStackTrace();
						return null;
					}
					entry.binaryContent = entryBinaryContent;
					
					if (jarEntry != null) {
						entry.certificates = jarEntry.getCertificates();
					}
				}
			}
		} finally {
			if (is != null) {
				try {
					is.close();
				} catch (IOException e) {}
			}
		}
		
		synchronized (resources) {
			ResourceCacheEntry entry2 = resources.get(name);
			if (entry2 == null) {
				resources.put(name, entry);
			} else {
				entry = entry2;
			}
		}
        return entry;
	}
	@Override
	public URL findResource(final String name) {
//		System.out.println("findResource("+name+")");
		ResourceCacheEntry entry = resources.get(name);
		if (entry == null) {
			entry = AccessController.doPrivileged(new PrivilegedAction<ResourceCacheEntry>() {
				public ResourceCacheEntry run() {
					return findResource0(name, name);
				}
			}, acc);
			if (entry != null) {
				resources.put(name, entry);
			}
		}
		if (entry != null) {
			return entry.getUrl();
		} else {
			return super.findResource(name);
		}
	}

	@Override
	public Enumeration<URL> findResources(String name) throws IOException {
		LinkedHashSet<URL> res = new LinkedHashSet<URL>();
		for (int i = 0; i < folders.length; i++) {
			String fullPath = folders[i].getAbsolutePath() + "/" + name;
			File file = new File(fullPath);
			try {
				file = file.getCanonicalFile();
			} catch (IOException e) {
				// Ignore
				e.printStackTrace();
			}
			if (file.exists()) {
				res.add(file.toURI().toURL());
			}
		}
		
		synchronized (jars) {
			for (int i = 0; i < jars.length; i++) {
				JarEntry jarEntry = jars[i].getJarEntry(name);
				if (jarEntry != null) {
					try {
						String entryUrlInJar = "jar:" + jarFiles[i].toURI().toURL().toString() + "!/" + name;
						res.add(new URL(entryUrlInJar));
					} catch (MalformedURLException e) {
						// Ignore
						e.printStackTrace();// Just for debug
					}
				}
			}
		}
		
		final Iterator<URL> iter = res.iterator();
		return new Enumeration<URL>() {

			public boolean hasMoreElements() {
				return iter.hasNext();
			}

			public URL nextElement() {
				return iter.next();
			}
		};
	}

//	@Override
//	protected PermissionCollection getPermissions(CodeSource codesource) {
//		System.out.println("Get Permissions... Delegate...");
//		return super.getPermissions(codesource);
//	}
//
//	@Override
////	Overridden because of delegation mecanism
//	public Class<?> loadClass(String name) throws ClassNotFoundException {
//		System.out.println("Load Class ("+name+")... Delegate...");
//		return super.loadClass(name);
//	}
//
//	@Override
////	Overridden because of delegation mecanism
//	protected synchronized Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
//		System.out.println("Load Class ("+name+", "+resolve+")... Delegate...");
//		return super.loadClass(name, resolve);
//	}
//
//	@Override
////	Overridden because of delegation mecanism
//	public URL getResource(String name) {
//		System.out.println("Get Resource ("+name+")... Delegate...");
//		return super.getResource(name);
//	}
//
//	@Override
////	Overridden because of delegation mecanism
//	public InputStream getResourceAsStream(String name) {
//		System.out.println("Get Resource as Stream ("+name+")... Delegate...");
//		return super.getResourceAsStream(name);
//	}
	
	public void dispose() {
		
		
		// TODO, but need an additional org.apache.catalina.loader.JdbcLeakPrevention class
//        // De-register any remaining JDBC drivers
//        clearReferencesJdbc();

//		// Stop any threads the web application started
//		clearReferencesThreads();
		
//        // Check for leaks triggered by ThreadLocals loaded by this class loader
//        checkThreadLocalsForLeaks();

//        // Clear RMI Targets loaded by this class loader
//        clearReferencesRmiTargets();

//		// TODO this seems interesting...
//        // Null out any static or final fields from loaded classes,
//        // as a workaround for apparent garbage collection bugs
//        if (clearReferencesStatic) {
//            clearReferencesStaticFinal();
//        }

//         // Clear the IntrospectionUtils cache.
//        IntrospectionUtils.clear();

//        // Clear the classloader reference in common-logging
//        if (clearReferencesLogFactoryRelease) {
////            org.apache.juli.logging.LogFactory.release(this);
//        }

//        // Clear the resource bundle cache
//        // This shouldn't be necessary, the cache uses weak references but
//        // it has caused leaks. Oddly, using the leak detection code in
//        // standard host allows the class loader to be GC'd. This has been seen
//        // on Sun but not IBM JREs. Maybe a bug in Sun's GC impl?
//        clearReferencesResourceBundles();

//		System.out.println("Dispose ClassLoader "+jarFiles.length);
        // Clear the classloader reference in the VM's bean introspector
        java.beans.Introspector.flushCaches();
        

		// Close jars and clean Files
		for (int i = 0; i < folders.length; i++) {
			folders[i] = null;
		}
		for (int i = 0; i < jars.length; i++) {
			try {
				if (jars[i] != null) {
					jars[i].close();
				}
			} catch (IOException e) {
				// Ignore
				e.printStackTrace();
			}
			jars[i] = null;
		}
		
		for (ResourceCacheEntry rce : resources.values()) {
			if (rce.clazz != null) {
				rce.clazz = null;
			}
		}
		resources.clear();
		folders = null;
		jars = null;
		jarFiles = null;
	}
	
}
