// ***************************************************************************************
// File: locks.cpp
// Description: lock objects (lock and semaphore)
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;


NvLock::NvLock()
{
	sName = "";
	sOwner = "";
	sInfo = "";
	LastUseTime = time(NULL);
}


NvLock::~NvLock()
{
}


bool NvLock::Lock(string Owner, string Info)
{
	LastUseTime = time(NULL);
	CleanString(Owner);
	MakeUpper(Owner);
	bool RetVal = false;
	if(sOwner.length())
	{
		// The object is own by somebody
		// Check if it's by us
		if(sOwner == Owner)
			RetVal = true;	// The object exists is is owned by us
		else
		{
			// The lock is not owned by us
			RetVal = false;
			// We must add it the pending list if it's not ever inside
			if(mPendings.find(Owner) == mPendings.end())
			{
				mPendings[Owner] = 1;
				vPendings.push_back(Owner);
			}
		}

	}
	else
	{
		// The object is not owned by anybody
		// If the pending list is empty, we can lock it
		if(mPendings.size() == 0)
		{
			RetVal = true;
		}
		else
		{
			// We can lock it only if this is the first item of the pending list
			if(vPendings.front() == Owner)
			{
				RetVal = true;
				vPendings.pop_front();
				mPendings.erase(Owner);
			}
		}
	}

	// All OK
	if(RetVal)
	{
		sOwner = Owner;
		sInfo = Info;
	}
	return RetVal;
}


bool NvLock::Unlock(string Owner, bool Force)
{
	LastUseTime = time(NULL);
	if(!Force)
	{
		CleanString(Owner);
		MakeUpper(Owner);
		if(sOwner.length())
		{
			if(sOwner != Owner)
				return false;
		}
	}

	sOwner = "";
	sInfo = "";
	return true;
}

bool NvLock::Islocked()
{
	LastUseTime = time(NULL);
	if(!(sOwner.length()))
		return false;

	return true;
}




NvLockList::NvLockList()
{
}


NvLockList::~NvLockList()
{
	Clear();
}


void NvLockList::Clear()
{
	NV_LOCKLOCK(&ProtectList);
	map<string,void *>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvLock *Lock = (NvLock *)(pos->second);
		Lock->Unlock(Lock->sOwner);
		delete Lock;
	}
	m_mList.clear();
	NV_UNLOCKLOCK(&ProtectList);
}



// WaitTime  gives the maximum number of seconds to wait for (-1 means infinite and 0 means no wait)
bool NvLockList::Lock(string Name, string Owner, string Info, int WaitTime, int Speed)
{
	if(Speed < 1)
		Speed = 1;
	if(Speed > 10)
		Speed = 10;
	int MaxSleepTime = 1000/(Speed*Speed);
	int MinSleepTime = 11-Speed;
	int CheckEvery = MinSleepTime;
	time_t InitialTime = time(NULL);
	bool RetVal = false;
	while(1)
	{
		NvLock *Lock = NULL;
		NV_LOCKLOCK(&ProtectList);
		try
		{
			Lock = GetLock(Name, true);	// We create the lock if necessary
			if(Lock != NULL)
				RetVal = Lock->Lock(Owner, Info);
		}
		catch(...)
		{
			GetMain()->IncNumExceptions();
			GetMain()->Logs.Write("Exceptions", "Exception when locking a lock object", Name.c_str(), NV_LOG_ERROR);
			RetVal = false;
			NV_UNLOCKLOCK(&ProtectList);
			break;
		}
		NV_UNLOCKLOCK(&ProtectList);
		if(RetVal)
			break;
		if(WaitTime == 0)
				break;
		if(WaitTime != -1)
		{
			time_t ActualTime = time(NULL);
			if((int)(ActualTime - InitialTime) > WaitTime)
				break;
		}
		SLEEP(CheckEvery);

		// We first do the check very often but less and less while the time is growing
		CheckEvery += (CheckEvery/10)+1;
		if(CheckEvery > MaxSleepTime)		// When the check time reaches the limits, we check every limit milliseconds
			CheckEvery = MaxSleepTime;
	}
	return RetVal;
}



bool NvLockList::Unlock(string Name, string Owner, bool Force)
{
	NV_LOCKLOCK(&ProtectList);

	NvLock *Lock = NULL;
	bool RetVal = true;
	try
	{
		Lock = GetLock(Name);
		if(Lock != NULL)
			RetVal = Lock->Unlock(Owner, Force);
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		GetMain()->Logs.Write("Exceptions", "Exception when unlocking a lock object", Name.c_str(), NV_LOG_ERROR);
		RetVal = false;
	}
	NV_UNLOCKLOCK(&ProtectList);
	return RetVal;
}

bool NvLockList::Islocked(string Name, string &Owner, string& Info)
{
	bool RetVal=false;
	Owner = "";
	Info = "";
	NV_LOCKLOCK(&ProtectList);
	NvLock *Lock = GetLock(Name);
	if(Lock == NULL)
	{
		NV_UNLOCKLOCK(&ProtectList);
		return RetVal;
	}

	RetVal = Lock->Islocked();
	if(RetVal)
	{
		Owner = Lock->sOwner;
		Info = Lock->sInfo;
	}

	NV_UNLOCKLOCK(&ProtectList);
	return RetVal;
}


NvLock* NvLockList::GetLock(string Name, bool Create)
{
	NvLock *Lock = NULL;
	CleanString(Name);
	if(!Name.length())
		return NULL;
	MakeUpper(Name);
	RemoveOldLocks();
	if(m_mList.find(Name) != m_mList.end())
		Lock = (NvLock *)(m_mList[Name]);
	if(Lock == NULL)
	{
		if(Create)
			Lock = NVNEW NvLock;
		if(Lock != NULL)
		{
			Lock->sName = Name;
			m_mList[Name] = (void *)Lock; 
		}
	}
	return Lock;
}

void NvLockList::UnlockAll(string Owner)
{
	CleanString(Owner);
	if(!Owner.length())
		return;
	MakeUpper(Owner);
	NV_LOCKLOCK(&ProtectList);
	map<string,void *>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvLock *LockInList = (NvLock *)(pos->second);
		if(Owner == LockInList->sOwner)
			LockInList->Unlock(Owner);
		if(LockInList->mPendings.find(Owner) != LockInList->mPendings.end())
		{
			LockInList->mPendings.erase(Owner);
			LockInList->vPendings.remove(Owner);
		}
	}
	NV_UNLOCKLOCK(&ProtectList);
}


void NvLockList::RemoveOldLocks()
{
	// The ProtectList object must be locked when entering this function
	bool Modified = false;
	time_t ActualTime = time(NULL);
	map<string,void *>::iterator pos;
	vector<string> m_vLocksToClear;
	try
	{
		for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
		{
			NvLock *LockInList = (NvLock *)(pos->second);
			if(LockInList->sOwner.length())
				continue;
			if((int)(ActualTime - LockInList->LastUseTime) > 60)	// We remove locks unused since 60 seconds
			{
				delete LockInList;
				m_vLocksToClear.push_back(pos->first);
			}
		}
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		GetMain()->Logs.Write("Exceptions", "Exception when removing old locks", NULL, NV_LOG_ERROR);
	}
	int i;
	for(i=0; i<(int)(m_vLocksToClear.size()); i++)
		m_mList.erase(m_vLocksToClear[i]);
	m_vLocksToClear.clear();
}



bool NvLockList::GetInfo(string Name, NvObjTable *LockInfo)
{
	LockInfo->Clear();

	bool AllLocks = true;
	
	LockInfo->InsertColumn("NAME");
	LockInfo->InsertColumn("OWNER");
	LockInfo->InsertColumn("INFO");

	CleanString(Name);
	MakeUpper(Name);
	if(Name.length() && (Name[0] == '*'))
		Name = "";


	NV_LOCKLOCK(&ProtectList);
	RemoveOldLocks();
	int j;
	map<string,void *>::iterator pos;
	for(pos = m_mList.begin(), j=0; pos != m_mList.end(); ++pos)
	{
		NvLock *LockInList = (NvLock *)(pos->second);
		if(!(LockInList->sOwner.length()))
			continue;
		if(Name.length())
		{
			string NameToSearch = Name;
			if(NameToSearch[NameToSearch.length() - 1] == '*')
			{
				// Search for several values
				NameToSearch = NameToSearch.substr(0, NameToSearch.length() - 1);
				if(LockInList->sName.find(NameToSearch) != 0)
					continue;
			}
			else
			{
				// Search for a specific value
				if(NameToSearch != LockInList->sName)
					continue;
			}
		}
		LockInfo->InsertRow();
		LockInfo->InsertCellLine(j, 0, LockInList->sName);
		LockInfo->InsertCellLine(j, 1, LockInList->sOwner);
		LockInfo->InsertCellLine(j, 2, LockInList->sInfo);
		j++;
	}
	NV_UNLOCKLOCK(&ProtectList);

	// Sort on name
	LockInfo->Sort("NAME");
	return true;
}






NvSema::NvSema()
{
	sName = "";
	LastUseTime = time(NULL);
	MaxValue = 1;
}


NvSema::~NvSema()
{
	UnlockAll();
}


int NvSema::GetInUse()
{
	NV_LOCKLOCK(&ProtectSema);
	int RetVal = Owners.GetSize();
	NV_UNLOCKLOCK(&ProtectSema);
	return RetVal;
}


bool NvSema::Init(string Name, int Value)
{
	NV_LOCKLOCK(&ProtectSema);
	sName = Name;
	if(Value < 0)
		Value = 1;
	MaxValue = Value;
	SemaObject.Set(Value);
	NV_UNLOCKLOCK(&ProtectSema);

	return true;
}


void NvSema::SetLastUseTime()
{
	NV_LOCKLOCK(&ProtectSema);
	LastUseTime = time(NULL);
	NV_UNLOCKLOCK(&ProtectSema);
}


time_t NvSema::GetLastUseTime()
{
	time_t RetVal;
	NV_LOCKLOCK(&ProtectSema);
	RetVal = LastUseTime;
	NV_UNLOCKLOCK(&ProtectSema);

	return RetVal;
}


bool NvSema::Lock(string Owner)
{
	CleanString(Owner);
	MakeUpper(Owner);
	if(!Owner.length())
		return false;

	NV_LOCKLOCK(&ProtectSema);
	if(Owners.KeyExist(Owner))
	{
		// The lock is ever owned by the session
		// so we don't do anything
		NV_UNLOCKLOCK(&ProtectSema);
		return true;
	}
	Owners.SetValue(Owner.c_str(), "");
	NV_UNLOCKLOCK(&ProtectSema);

	// We can now lock the semaphore object
	SemaObject.Lock();
	return true;
}


bool NvSema::Unlock(string Owner)
{
	CleanString(Owner);
	MakeUpper(Owner);
	if(!Owner.length())
		return false;

	NV_LOCKLOCK(&ProtectSema);
	if(!Owners.KeyExist(Owner))
	{
		// The lock is not owned by the session
		// so we don't do anything
		NV_UNLOCKLOCK(&ProtectSema);
		return true;
	}
	SemaObject.Unlock();
	Owners.Remove(Owner);
	NV_UNLOCKLOCK(&ProtectSema);

	return true;
}


bool NvSema::UnlockAll()
{
	NV_LOCKLOCK(&ProtectSema);
	int i;
	for(i=0; i<Owners.GetSize(); i++)
		SemaObject.Unlock();
	Owners.RemoveAll();
	NV_UNLOCKLOCK(&ProtectSema);

	return true;
}


// This function unlocks the sema if owned by deleted sessions
void NvSema::CheckSessions(NvObjIndStringList *SessionIds)
{
	NV_LOCKLOCK(&ProtectSema);
	int i;
	NvObjStringList OwnersInSema;
	map<string,string>::iterator pos;
	for(pos = Owners.m_mList.begin(); pos != Owners.m_mList.end(); ++pos)
	{
		string OwnerInSema = pos->first;
		OwnersInSema.Add(OwnerInSema);
	}
	for(i=0; i<OwnersInSema.GetSize(); i++)
	{
		string OwnerInSema = OwnersInSema.GetValue(i);
		if(!SessionIds->KeyExist(OwnerInSema))
		{
			SemaObject.Unlock();
			Owners.Remove(OwnerInSema);
		}
	}
	NV_UNLOCKLOCK(&ProtectSema);
}


NvSemaList::NvSemaList()
{
}


NvSemaList::~NvSemaList()
{
	Clear();
}


void NvSemaList::Clear()
{
	NV_LOCKLOCK(&ProtectList);
	map<string,void *>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvSema *Sema = (NvSema *)(pos->second);
		delete Sema;
	}
	m_mList.clear();
	NV_UNLOCKLOCK(&ProtectList);
}



bool NvSemaList::Lock(string Name, string Owner, int Value)
{
	bool RetVal = false;
	NV_LOCKLOCK(&ProtectList);
	NvSema *Sema = GetSema(Name, true, Value);	// We create the lock if necessary

	if(Sema == NULL)
	{
		NV_UNLOCKLOCK(&ProtectList);
		return false;
	}
	NV_UNLOCKLOCK(&ProtectList);
	Sema->Lock(Owner);

	return true;
}



bool NvSemaList::Unlock(string Name, string Owner)
{
	NV_LOCKLOCK(&ProtectList);
	NvSema *Sema = GetSema(Name);
	if(Sema == NULL)
	{
		NV_UNLOCKLOCK(&ProtectList);
		return true;
	}

	bool RetVal = Sema->Unlock(Owner);
	NV_UNLOCKLOCK(&ProtectList);
	return RetVal;
}


NvSema* NvSemaList::GetSema(string Name, bool Create, int Value)
{
	if(Value <= 0)
		Value = 1;
	NvSema *Sema = NULL;
	CleanString(Name);
	if(!Name.length())
		return NULL;
	MakeUpper(Name);
	RemoveOldSemas();
	if(m_mList.find(Name) != m_mList.end())
		Sema = (NvSema *)(m_mList[Name]);
	if(Sema == NULL)
	{
		if(Create)
			Sema = NVNEW NvSema;
		if(Sema != NULL)
		{
			Sema->Init(Name, Value);
			m_mList[Name] = (void *)Sema; 
		}
	}
	if(Sema != NULL)
		Sema->SetLastUseTime();
	return Sema;
}

void NvSemaList::UnlockAll(string Owner)
{
	NV_LOCKLOCK(&ProtectList);
	map<string,void *>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvSema *LockInList = (NvSema *)(pos->second);
		LockInList->Unlock(Owner);
	}
	NV_UNLOCKLOCK(&ProtectList);
}


void NvSemaList::RemoveOldSemas()
{
	// The ProtectList object must be locked when entering this function
	bool Modified = false;
	time_t ActualTime = time(NULL);
	map<string,void *>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvSema *SemaInList = (NvSema *)(pos->second);
		if(SemaInList->GetInUse())
			continue;
		if((int)(ActualTime - SemaInList->GetLastUseTime()) > 60)	// We remove semas unused since 60 seconds
		{
			delete SemaInList;
			m_mList.erase(pos->first);
		}
	}
}



bool NvSemaList::GetInfo(string Name, NvObjTable *SemaInfo)
{
	SemaInfo->Clear();
	
	// Do a check sessions first
	CheckSessions();

	bool AllSemas = true;
	
	SemaInfo->InsertColumn("NAME");
	SemaInfo->InsertColumn("INUSE");
	SemaInfo->InsertColumn("VALUE");

	CleanString(Name);
	MakeUpper(Name);
	if(Name.length() && (Name[0] == '*'))
		Name = "";


	NV_LOCKLOCK(&ProtectList);
	RemoveOldSemas();
	int j;
	map<string,void *>::iterator pos;
	for(pos = m_mList.begin(), j=0; pos != m_mList.end(); ++pos)
	{
		NvSema *SemaInList = (NvSema *)(pos->second);
		if(Name.length())
		{
			string NameToSearch = Name;
			if(NameToSearch[NameToSearch.length() - 1] == '*')
			{
				// Search for several values
				NameToSearch = NameToSearch.substr(0, NameToSearch.length() - 1);
				if(SemaInList->sName.find(NameToSearch) != 0)
					continue;
			}
			else
			{
				// Search for a specific value
				if(NameToSearch != SemaInList->sName)
					continue;
			}
		}
		SemaInfo->InsertRow();
		SemaInfo->InsertCellLine(j, 0, SemaInList->sName);
		char Buffer[50];
		sprintf(Buffer, "%d", SemaInList->GetInUse());
		SemaInfo->InsertCellLine(j, 1, Buffer);
		sprintf(Buffer, "%d", SemaInList->MaxValue);
		SemaInfo->InsertCellLine(j, 2, Buffer);
		j++;
	}
	NV_UNLOCKLOCK(&ProtectList);

	// Sort ton name
	SemaInfo->Sort("NAME");
	return true;
}




// This function unlocks all semas owned by deleted sessions
void NvSemaList::CheckSessions()
{
	NvObjStringList SessionIds;
	
	// We get the actual session Id list
	// This must be done outside the ProtectList because otherwise
	// this may generates a deadlock between ProtectList and SessionList
	GetMain()->Sessions.List(&SessionIds);
	NV_LOCKLOCK(&ProtectList);
	RemoveOldSemas();
	if(!m_mList.size())
	{
		NV_UNLOCKLOCK(&ProtectList);
		return;
	}
	NvObjIndStringList SessionIdsHere;
	int i;
	for( i=0; i<SessionIds.GetSize(); i++)
	{
		string Owner = SessionIds.GetValue(i);
		CleanString(Owner);
		MakeUpper(Owner);
		SessionIdsHere.SetValue(Owner, "");
	}
	map<string,void *>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvSema *SemaInList = (NvSema *)(pos->second);
		SemaInList->CheckSessions(&SessionIdsHere);
	}
	RemoveOldSemas();
	NV_UNLOCKLOCK(&ProtectList);
}



// Removes a semaphore if its not in use
bool NvSemaList::Remove(string Name)
{
	CleanString(Name);
	if(!Name.length())
		return true;
	MakeUpper(Name);

	// Do a check sessions first
	CheckSessions();

	NV_LOCKLOCK(&ProtectList);
	bool bFound = false;
	bool bDeleted = false;
	map<string,void *>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvSema *SemaInList = (NvSema *)(pos->second);
		if(SemaInList->sName == Name)
		{
			bFound = true;
			// We have found it
			if(!(SemaInList->GetInUse()))
			{
				delete SemaInList;
				m_mList.erase(pos->first);
				bDeleted = true;
			}
			break;
		}
	}
	NV_UNLOCKLOCK(&ProtectList);

	if(!bFound)
		return true;	// Not found semaphore

	if(!bDeleted)
		return false;	// Semaphore found but cannot be deleted (probably in use)
	
	return true;
}
