// ***************************************************************************************
// File: service.cpp
// Description: nirva external service interface
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;



/////////////////////////////////////////////////////////////////////////////
// NvServiceList

NvServiceList::NvServiceList()
{
	// We create the default service
}

NvServiceList::~NvServiceList()
{
	Clear();
}



// No lock because used only at exit time from nvs.cpp
void NvServiceList::Clear()
{
	map<string,void *>::iterator pos;
	for(pos = m_mServices.begin(); pos != m_mServices.end(); ++pos)
	{
		NvService *Service = (NvService *)(pos->second);
		if(Service != NULL)
			Service->Stop();
	}
	for(pos = m_mServices.begin(); pos != m_mServices.end(); ++pos)
	{
		NvService *Service = (NvService *)(pos->second);
		if(Service != NULL)
			delete Service;
	}
	m_mServices.clear();
}



void NvServiceList::ClearRegistryCache()
{
	NV_LOCKOBJECT(this);
	map<string,void *>::iterator pos;
	for(pos = m_mServices.begin(); pos != m_mServices.end(); ++pos)
	{
		NvService *Service = (NvService *)(pos->second);
		if(Service != NULL)
			if(!Service->bInInstall)
				Service->SrvRegistry.Clear();
	}
	NV_UNLOCKOBJECT(this);
}

// Load
// Loads the service list from the system registry
// This function is called only one time when starting the server so there is no need of locking
void NvServiceList::Load()
{
	// load the service table
	// The service table is stored in a table object named "services" in the
	// system registry key system.services
	NvContainer Container;
	if(GetMain()->SystemRegistry.GetKey("system.services", &Container, NV_CONTAINER_COPY_NO_APPEND, "services", false, true))
	{
		NvObj *Object = Container.GetObject("services");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_TABLE))
		{
			NvObjTable *ServiceTable = (NvObjTable*)Object;
			int index_Name = ServiceTable->GetColumnIndex("Name");
			int index_Desc = ServiceTable->GetColumnIndex("Description");
			int index_Path = ServiceTable->GetColumnIndex("Path");
			int index_Multithread = ServiceTable->GetColumnIndex("Multithread");
			int index_LoadOnStart = ServiceTable->GetColumnIndex("LoadOnStart");
			int index_Enable = ServiceTable->GetColumnIndex("Enable");
			int index_Java = ServiceTable->GetColumnIndex("Java");
			int index_Language = ServiceTable->GetColumnIndex("Language");	// Not used any more but kept for comaptibility
			int row_index;
			for(row_index=0; row_index<ServiceTable->GetNumRow(); row_index++)
			{
				// Get service parameters
				string ServiceName = ServiceTable->GetCellLine(row_index, index_Name);
				NvService::CleanSrvName(ServiceName);
				if(!ServiceName.length()) // The service is identified by its name
					continue;
				if(InternalGetService(ServiceName.c_str()) != NULL)	// Check doubles
					continue;
				string ServiceDesc = ServiceTable->GetCellLine(row_index, index_Desc);
				string ServicePath = ServiceTable->GetCellLine(row_index, index_Path);
				string ServiceMultithread = ServiceTable->GetCellLine(row_index, index_Multithread);
				string ServiceLoadOnStart = ServiceTable->GetCellLine(row_index, index_LoadOnStart);
				string ServiceEnable = ServiceTable->GetCellLine(row_index, index_Enable);
				string ServiceJava = ServiceTable->GetCellLine(row_index, index_Java);
				string ServiceLanguage = "";
				if(index_Language != -1)
					ServiceLanguage = ServiceTable->GetCellLine(row_index, index_Language);
				bool bServiceMultithread = true;
				if(ServiceMultithread.length() && (atoi(ServiceMultithread.c_str()) == 0))
					bServiceMultithread = false;
				bool bServiceLoadOnStart = false;
				if(ServiceLoadOnStart.length() && (atoi(ServiceLoadOnStart.c_str()) == 1))
					bServiceLoadOnStart = true;
				bool bServiceEnabled = true;
				if(ServiceEnable.length() && (atoi(ServiceEnable.c_str()) == 0))
					bServiceEnabled = false;
				int Language = LANGUAGE_CPP;
				if(ServiceLanguage.length())
				{
					if(ServiceLanguage == "JAVA")
						Language = LANGUAGE_JAVA;
					if(ServiceLanguage == "DOTNET")
						Language = LANGUAGE_DOTNET;
				}
				else
				{
					// Old service list, we use the Java column
					if(ServiceJava.length() && (atoi(ServiceJava.c_str()) == 1))
						Language = LANGUAGE_JAVA;
				}

				// Create the service in service list
				if(!CreateService(ServiceName.c_str(), ServiceDesc.c_str(), ServicePath.c_str(), bServiceMultithread, bServiceLoadOnStart, Language, true))
					bServiceEnabled = false;

				if(bServiceEnabled)
					EnableService(ServiceName.c_str(), START_MODE_NOSTART, true);
			}
		}
	}

	// Saves the list in case of changes
	Save();
}


// Saves the service list to the system registry
// This function is called internally so its not locked
bool NvServiceList::Save()
{
	// First create the registry key "system.services" if it doesn't exist
	GetMain()->SystemRegistry.SetKey("system.services", NULL);

	NvContainer Container;
	if(!GetMain()->SystemRegistry.GetKey("system.services", &Container, NV_CONTAINER_COPY_NO_APPEND, NULL, false, true))
		return false; // cannot get the key. This should not occure here

	// Creates the service object if it doesn't exist
	Container.CreateObject("services", NV_OBJTYPE_TABLE);
	NvObj *Object = Container.GetObject("services");
	if(Object == NULL)
		return false; // This should not occur here

	if(Object->GetType() != NV_OBJTYPE_TABLE)
	{
		Container.RemoveObject("services");
		Container.CreateObject("services", NV_OBJTYPE_TABLE);
		Object = Container.GetObject("services");
		if(Object == NULL)
			return false; // This should not occur here
	}

	NvObjTable *ServiceTable = (NvObjTable*)Object;
	ServiceTable->Clear();
	ServiceTable->InsertColumn("Name");
	ServiceTable->InsertColumn("Description");
	ServiceTable->InsertColumn("Path");
	ServiceTable->InsertColumn("Multithread");
	ServiceTable->InsertColumn("LoadOnStart");
	ServiceTable->InsertColumn("Enable");
	ServiceTable->InsertColumn("Java");	// Not used any more but kept for compatibility
	ServiceTable->InsertColumn("Language");

	int index_Name = ServiceTable->GetColumnIndex("Name");
	int index_Desc = ServiceTable->GetColumnIndex("Description");
	int index_Path = ServiceTable->GetColumnIndex("Path");
	int index_Multithread = ServiceTable->GetColumnIndex("Multithread");
	int index_LoadOnStart = ServiceTable->GetColumnIndex("LoadOnStart");
	int index_Enable = ServiceTable->GetColumnIndex("Enable");
	int index_Java = ServiceTable->GetColumnIndex("Java");
	int index_Language = ServiceTable->GetColumnIndex("Language");

	int j;
	map<string,void *>::iterator pos;
	for(pos = m_mServices.begin(), j=0; pos != m_mServices.end(); ++pos)
	{
		NvService *Service = (NvService *)(pos->second);
		if(Service == NULL)
			continue;

		if(!Service->GetName().length())
			continue;

		if(!Service->GetPath().length())
			continue;

		ServiceTable->InsertRow();
		ServiceTable->InsertCellLine(j, index_Name, Service->GetName().c_str());
		ServiceTable->InsertCellLine(j, index_Desc, Service->GetDesc().c_str());
		ServiceTable->InsertCellLine(j, index_Path, Service->GetPath().c_str());
		if(Service->m_bMultithread)
			ServiceTable->InsertCellLine(j, index_Multithread, "1");
		else
			ServiceTable->InsertCellLine(j, index_Multithread, "0");
		if(Service->m_bLoadOnStart)
			ServiceTable->InsertCellLine(j, index_LoadOnStart, "1");
		else
			ServiceTable->InsertCellLine(j, index_LoadOnStart, "0");
		if(Service->IsEnable())
			ServiceTable->InsertCellLine(j, index_Enable, "1");
		else
			ServiceTable->InsertCellLine(j, index_Enable, "0");
		if(Service->m_iLanguage == LANGUAGE_JAVA)
			ServiceTable->InsertCellLine(j, index_Java, "1");
		else
			ServiceTable->InsertCellLine(j, index_Java, "0");
		if(Service->m_iLanguage == LANGUAGE_JAVA)
			ServiceTable->InsertCellLine(j, index_Language, "JAVA");
		else if(Service->m_iLanguage == LANGUAGE_DOTNET)
			ServiceTable->InsertCellLine(j, index_Language, "DOTNET");
		else
			ServiceTable->InsertCellLine(j, index_Language, "C++");
		j++;
	}
	
	if(!GetMain()->SystemRegistry.SetKey("system.services", &Container))
		return false;

	return true;
}


// Starts services having the load on start flag set
// This function is called only one time when starting the server so there is no need of locking
void NvServiceList::LoadOnStart()
{
	map<string,void *>::iterator pos;
	for(pos = m_mServices.begin(); pos != m_mServices.end(); ++pos)
	{
		NvService *Service = (NvService *)(pos->second);
		if(Service == NULL)
			continue;
	
		if(Service->m_bLoadOnStart)
			Service->Start();
	}

}


NvService* NvServiceList::InternalGetService(const char *Name)
{
	if(Name == NULL)
		return NULL;

	string NameHere = Name;
	NvService::CleanSrvName(NameHere);

	if(!NameHere.length())
		return NULL;

	NvService *FoundService = NULL;
	if(m_mServices.find(NameHere) != m_mServices.end())
		FoundService = (NvService *)(m_mServices[NameHere]);

	return FoundService;
}


NvService* NvServiceList::GetService(const char *Name, bool LockIt, bool CheckEnable)
{
	if(Name == NULL)
		return NULL;
	NvService *ReturnedService = NULL;
	NV_LOCKOBJECT(this);
	ReturnedService = InternalGetService(Name);

	if(CheckEnable && (ReturnedService != NULL) && (!ReturnedService->IsEnable()))
		ReturnedService = NULL;

	if(LockIt && (ReturnedService != NULL))
		if(!ReturnedService->IncInuse())
			ReturnedService = NULL;			// The service is disabled		
	NV_UNLOCKOBJECT(this);
	return ReturnedService;
}


bool NvServiceList::EnableService(const char *Name, int StartIt, bool FromLoad)
{
	if(Name == NULL)
		return false;
	NvService *ReturnedService = NULL;
	NV_LOCKOBJECT(this);
	ReturnedService = InternalGetService(Name);

	if(ReturnedService == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(ReturnedService->bInInstall)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	// We first enable the service
	bool RetVal = ReturnedService->Enable(START_MODE_NOSTART);
	if(!FromLoad)
		if(RetVal)
			Save();

	// Then if necessary, we start it
	if(RetVal && ((StartIt == START_MODE_START) || ((StartIt == START_MODE_FLAG) && ReturnedService->m_bLoadOnStart)))
	{
		NV_UNLOCKOBJECT(this);
		GetService(Name);
		RetVal = ReturnedService->Enable(StartIt);
		FreeService(Name);
		NV_LOCKOBJECT(this);
	}

	if(!FromLoad)
		if(RetVal)
			Save();

	NV_UNLOCKOBJECT(this);
	return RetVal;
}



bool NvServiceList::CheckRequiredForApplications(const char *Name)
{
	if(Name == NULL)
		return true;
	NvService *ReturnedService = NULL;
	NV_LOCKOBJECT(this);
	ReturnedService = InternalGetService(Name);

	if(ReturnedService == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}
	if(ReturnedService->IsEnable() && ReturnedService->CheckRequiredForApplications())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	NV_UNLOCKOBJECT(this);
	return true;
}


bool NvServiceList::DisableService(const char *Name, const char *SessionId, int WaitTime)
{
	if(Name == NULL)
		return false;
	NvService *ReturnedService = NULL;
	NV_LOCKOBJECT(this);
	ReturnedService = InternalGetService(Name);

	if(ReturnedService == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}
	// If the service is required for an application, we cannot stop it
	if(ReturnedService->IsEnable() && ReturnedService->CheckRequiredForApplications())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	ReturnedService->SetDisableFlag();
	NV_UNLOCKOBJECT(this);
	bool RetVal = ReturnedService->Disable(SessionId, WaitTime);
	NV_LOCKOBJECT(this);
	if(RetVal)
		Save();
	NV_UNLOCKOBJECT(this);
	return RetVal;
}



bool NvServiceList::CreateService(const char *Name, const char *Desc, const char *Path, bool Multithread, bool LoadOnStart, int Language, bool FromLoad)
{
	if(Name == NULL)
		return false;

	string NameHere = Name;
	NvService::CleanSrvName(NameHere);

	if(!NameHere.length())
		return false;
	
	string PathHere = "";
	if(Path != NULL)
		PathHere = Path;
	CleanString(PathHere);

	NV_LOCKOBJECT(this);
	
	// Test if the service exists
	NvService *Service = InternalGetService(NameHere.c_str());
	bool bNewService = true;
	if(Service != NULL)
	{
		if(Service->IsEnable())
		{
			NV_UNLOCKOBJECT(this);
			return false;	// The service is actually running, we cannot change its parameters
		}
		if(Service->bInInstall)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		bNewService = false;
	}
	else
	{
		Service = NVNEW NvService;
		if(Service == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	bool RetVal = Service->SetParameters(NameHere.c_str(), Desc, PathHere.c_str(), Multithread, LoadOnStart, Language);
	if(!FromLoad)
	{
		if(!RetVal)
		{
			if(bNewService)
				delete Service;
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	if(bNewService)
	{
		m_mServices[Service->GetName()] = (void *)Service;
		// Load logs
		Service->Logs.m_sServiceName = Service->m_sName;
		Service->Logs.Owner = (void *)Service;
		Service->Logs.Load();
	}

	if(!FromLoad)
	{
		if(!Save())
		{
			if(bNewService)
			{
				m_mServices.erase(Service->GetName());
				delete Service;
			}
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}
	
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvServiceList::CreateSourceSkeleton(const char *Name, bool WithSessions, int Language, const char *SessionId)
{
	if(Name == NULL)
		return false;

	bool Found = true;
	NvService *Service = NULL;
	NV_LOCKOBJECT(this);
	Service = InternalGetService(Name);
	if(Service == NULL)
	{
		// The service has not been found so create a temporary service
		Service = NVNEW NvService;
		if(Service == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		string NameHere = Name;
		CleanString(NameHere);
		MakeUpper(NameHere);
		if(!NameHere.length())
		{
			delete Service;
			NV_UNLOCKOBJECT(this);
			return false;
		}
		Service->m_sName = NameHere;
		Service->m_iLanguage = Language;
		Found = false;
	}

	if(Service->bInInstall)
	{
		if(!Found)
			delete Service;
		NV_UNLOCKOBJECT(this);
		return false;
	}
	bool RetVal = Service->CreateSourceSkeleton(WithSessions, Language, SessionId);

	if(!Found)
		delete Service;

	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvServiceList::Install(const char *Name, const char *PackageFile, bool bMountIt, string& Info)
{
	if(Name == NULL)
		return false;

	bool Found = true;
	NvService *Service = NULL;
	NV_LOCKOBJECT(this);
	Service = InternalGetService(Name);
	string NameHere = Name;
	NvService::CleanSrvName(NameHere);
	if(Service == NULL)
	{
		// The service has not been found so create a temporary service
		Service = NVNEW NvService;
		if(Service == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(!NameHere.length())
		{
			delete Service;
			NV_UNLOCKOBJECT(this);
			return false;
		}
		Service->m_sName = NameHere;
		Found = false;
	}
	else
	{
		if(Service->bInInstall)
		{
			Info = "The service is ever in install mode";
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(Service->IsEnable())
		{
			Info = "The service is running";
			NV_UNLOCKOBJECT(this);
			return false;	// The service is actually running, we cannot install it
		}
	}

	// Empty the xslt cache
	NvXml::EmptyCache();

	bool RetVal = Service->Install(PackageFile, Info);

	NV_UNLOCKOBJECT(this);

	if(bMountIt && RetVal && (!Found))
	{
		// We must get the service language from the desc file
		NV_LOCKOBJECT(this);
		Service->LoadInfo();
		NvObjIndStringList *IndStringList = (NvObjIndStringList *)(Service->m_Info.GetObject("SETTINGS"));
		int Language = LANGUAGE_CPP;
		if(IndStringList != NULL)
		{
			string sLanguage = IndStringList->GetValue("LANGUAGE");
			CleanString(sLanguage, true);
			MakeUpper(sLanguage);
			if(sLanguage == "JAVA")
				Language = LANGUAGE_JAVA;
			if(sLanguage == "DOTNET")
				Language = LANGUAGE_DOTNET;
		}
		NV_UNLOCKOBJECT(this);
		RetVal = CreateService(NameHere.c_str(), "", "", true, false, Language, false);
	}

	return RetVal;
}


bool NvServiceList::Package(const char *Name, const char *DescriptionFile, const char *PackageFile, string& Info)
{
	if(Name == NULL)
		return false;

	bool Found = true;
	NvService *Service = NULL;
	NV_LOCKOBJECT(this);
	Service = InternalGetService(Name);
	if(Service == NULL)
	{
		// The service has not been found so create a temporary service
		Service = NVNEW NvService;
		if(Service == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		string NameHere = Name;
		CleanString(NameHere);
		MakeUpper(NameHere);
		if(!NameHere.length())
		{
			delete Service;
			NV_UNLOCKOBJECT(this);
			return false;
		}
		Service->m_sName = NameHere;
		Found = false;
	}

	if(Service->bInInstall)
	{
		if(!Found)
			delete Service;
		NV_UNLOCKOBJECT(this);
		return false;
	}
	bool RetVal = Service->Package(DescriptionFile, PackageFile, Info);

	if(!Found)
		delete Service;

	NV_UNLOCKOBJECT(this);
	return RetVal;
}




void NvServiceList::FreeService(const char *Name)
{
	NvService *FoundService = NULL;
	NV_LOCKOBJECT(this);
	FoundService = InternalGetService(Name);
	if(FoundService != NULL)
		FoundService->DecInUse();
	NV_UNLOCKOBJECT(this);
}


bool NvServiceList::RemoveService(const char *Name, const char *SessionId, int WaitTime)
{
	NvService *FoundService = NULL;
	NV_LOCKOBJECT(this);
	string NameHere = Name;
	NvService::CleanSrvName(NameHere);
	FoundService = InternalGetService(NameHere.c_str());
	if(FoundService != NULL)
	{
		if(FoundService->bInInstall)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		FoundService->SetDisableFlag();
		NV_UNLOCKOBJECT(this);
		if(!FoundService->Disable(SessionId, WaitTime))
			return false;
		NV_LOCKOBJECT(this);
		m_mServices.erase(NameHere);
		if(!Save())
		{
			m_mServices[NameHere] = ((void *)FoundService);
			NV_UNLOCKOBJECT(this);
			return false;
		}
		delete FoundService;

	}
	NV_UNLOCKOBJECT(this);
	return true;
}


bool NvServiceList::ServiceExist(const char *Name)
{
	NvService *FoundService = NULL;
	NV_LOCKOBJECT(this);
	FoundService = InternalGetService(Name);
	NV_UNLOCKOBJECT(this);
	if(FoundService == NULL)
		return false;

	return true;
}


bool NvServiceList::IsServiceEnable(const char *Name)
{
	NvService *FoundService = NULL;
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	FoundService = InternalGetService(Name);
	if(FoundService != NULL)
		if(FoundService->IsEnable())
			RetVal = true;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}



bool NvServiceList::IsRenderer(const char *Name)
{
	NvService *FoundService = NULL;
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	FoundService = InternalGetService(Name);
	if(FoundService != NULL)
	{
		if(FoundService->IsEnable() && FoundService->IsRenderer())
				RetVal = true;
	}
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvServiceList::GetSessions(const char *Name, NvObjStringList *ConnectedSessions)
{
	ConnectedSessions->RemoveAll();
	NvService *FoundService = NULL;
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	FoundService = InternalGetService(Name);
	if(FoundService != NULL)
		if(FoundService->GetSessions(ConnectedSessions))
			RetVal = true;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}



bool NvServiceList::GetInfo(const char *Name, NvObjTable *ServiceInfo)
{
	ServiceInfo->Clear();

	bool AllServices = true;

	NV_LOCKOBJECT(this);

	NvService *FoundService = NULL;
	if((Name != NULL) && strlen(Name))
	{
		// There is a request on a specific service
		FoundService = InternalGetService(Name);
		if(FoundService == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		AllServices = false;
	}


	ServiceInfo->InsertColumn("NAME");
	ServiceInfo->InsertColumn("DESCRIPTION");
	ServiceInfo->InsertColumn("PATH");
	ServiceInfo->InsertColumn("STATUS");
	ServiceInfo->InsertColumn("MULTITHREAD");
	ServiceInfo->InsertColumn("LOAD_ON_START");
	ServiceInfo->InsertColumn("LANGUAGE");
	ServiceInfo->InsertColumn("NUM_SESSIONS");
	ServiceInfo->InsertColumn("REQUIRED_FOR_APPLICATIONS");

	if(AllServices)
	{
		int j;
		map<string,void *>::iterator pos;
		for(pos = m_mServices.begin(),j=0; pos != m_mServices.end(); ++pos)
		{
			NvService *Service = (NvService *)(pos->second);
			if(Service == NULL)
				continue;
			ServiceInfo->InsertRow();
			ServiceInfo->InsertCellLine(j, 0, Service->GetName());
			ServiceInfo->InsertCellLine(j, 1, Service->GetDesc());
			ServiceInfo->InsertCellLine(j, 2, Service->GetPath());
			if(Service->IsEnable())
				ServiceInfo->InsertCellLine(j, 3, "RUNNING");
			else
				ServiceInfo->InsertCellLine(j, 3, "STOPPED");
			if(Service->m_bMultithread)
				ServiceInfo->InsertCellLine(j, 4, "YES");
			else
				ServiceInfo->InsertCellLine(j, 4, "NO");
			if(Service->m_bLoadOnStart)
				ServiceInfo->InsertCellLine(j, 5, "YES");
			else
				ServiceInfo->InsertCellLine(j, 5, "NO");
			if(Service->m_iLanguage == LANGUAGE_JAVA)
				ServiceInfo->InsertCellLine(j, 6, "JAVA");
			else if(Service->m_iLanguage == LANGUAGE_DOTNET)
				ServiceInfo->InsertCellLine(j, 6, "DOTNET");
			else
				ServiceInfo->InsertCellLine(j, 6, "C++");
			char Buffer[20];
			sprintf(Buffer, "%d", Service->Sessions.GetSize());
			ServiceInfo->InsertCellLine(j, 7, Buffer);
			string SrvName = Service->GetName();
			ServiceInfo->InsertCellLine(j, 8, GetMain()->Applications.GetAppForRequiredService(SrvName));
			j++;
		}
	}
	else
	{
		ServiceInfo->InsertRow();
		ServiceInfo->InsertCellLine(0, 0, FoundService->GetName());
		ServiceInfo->InsertCellLine(0, 1, FoundService->GetDesc());
		ServiceInfo->InsertCellLine(0, 2, FoundService->GetPath());
		if(FoundService->IsEnable())
			ServiceInfo->InsertCellLine(0, 3, "RUNNING");
		else
			ServiceInfo->InsertCellLine(0, 3, "STOPPED");
		if(FoundService->m_bMultithread)
			ServiceInfo->InsertCellLine(0, 4, "YES");
		else
			ServiceInfo->InsertCellLine(0, 4, "NO");
		if(FoundService->m_bLoadOnStart)
			ServiceInfo->InsertCellLine(0, 5, "YES");
		else
			ServiceInfo->InsertCellLine(0, 5, "NO");
		if(FoundService->m_iLanguage == LANGUAGE_JAVA)
			ServiceInfo->InsertCellLine(0, 6, "JAVA");
		else if(FoundService->m_iLanguage == LANGUAGE_DOTNET)
			ServiceInfo->InsertCellLine(0, 6, "DOTNET");
		else
			ServiceInfo->InsertCellLine(0, 6, "C++");
		char Buffer[20];
		sprintf(Buffer, "%d", FoundService->Sessions.GetSize());
		ServiceInfo->InsertCellLine(0, 7, Buffer);
		string SrvName = FoundService->GetName();
		ServiceInfo->InsertCellLine(0, 8, GetMain()->Applications.GetAppForRequiredService(SrvName));
	}
	NV_UNLOCKOBJECT(this);

	return true;
}



bool NvServiceList::GetInfoEx(const char *Name, NvObjIndStringList *ServiceInfo)
{
	if(ServiceInfo == NULL)
		return false;

	ServiceInfo->RemoveAll();

	if(Name == NULL)
		return false;

	if(!strlen(Name))
		return false;

	NV_LOCKOBJECT(this);

	NvService *FoundService = NULL;
	
	// Try to get the service
	FoundService = InternalGetService(Name);
	if(FoundService == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	
	// We must load now the description file if it exists
	FoundService->LoadInfo();

	NvObjIndStringList *IndStringList = (NvObjIndStringList *)(FoundService->m_Info.GetObject("INFO"));

	if(IndStringList == NULL)
	{
		NV_UNLOCKOBJECT(this);		// Nothing to return
		return true;
	}

	ServiceInfo->CopyFrom(IndStringList);
	NV_UNLOCKOBJECT(this);

	return true;
}


bool NvServiceList::GetPermissions(NvObjTable *ServicePermissions)
{
	NV_LOCKOBJECT(this);

	map<string,void *>::iterator pos;
	for(pos = m_mServices.begin(); pos != m_mServices.end(); ++pos)
	{
		NvService *Service = (NvService *)(pos->second);
		if(Service == NULL)
			continue;

		NvObjIndStringList *IndStringList = (NvObjIndStringList *)(Service->m_Info.GetObject("PERMISSIONS"));
		if(IndStringList != NULL)
		{
			map<string,string>::iterator pos;
			string ServiceName = Service->GetName();
			CleanString(ServiceName);
			MakeUpper(ServiceName);
			ServiceName += ":";
			for(pos = IndStringList->m_mList.begin(); pos != IndStringList->m_mList.end(); ++pos)
			{
				string KeyToAdd = ServiceName;
				string KeyFromList = pos->first.c_str();
				CleanString(KeyFromList);
				MakeUpper(KeyFromList);
				if(!KeyFromList.length())
					continue;
				string ValueFromList = pos->second.c_str();
				CleanString(ValueFromList);
				if(!ValueFromList.length())
					continue;
				KeyToAdd += KeyFromList;
				ServicePermissions->InsertRow();
				ServicePermissions->InsertCellLine(-1, 0, KeyToAdd);
				ServicePermissions->InsertCellLine(-1, 1, ValueFromList);
			}
		}
	}

	NV_UNLOCKOBJECT(this);
	return true;
}

/////////////////////////////////////////////////////////////////////////////
// NvService

NvService::NvService()
{
	Initialized = false;
	Started = false;
	InUse = 0;
	m_sName = "";
	m_sDesc = "";
	m_sPath = ""; 
	m_bIsRenderer = false;
	m_sRegistryDir = ""; 
	m_sDocDir = ""; 
	m_sFileDir = ""; 
	m_sLogDir = ""; 
	m_sProcDir = ""; 
	m_sBinDir = ""; 
	m_sWrootDir = ""; 
	hLib = NULL;
	LibCommand = NULL;
	LibRender = NULL;
	m_bEnable = false;
	m_bMultithread = true;
	m_bLoadOnStart = false;
	m_iLanguage = LANGUAGE_CPP;
	LastSessionId = "";
	SrvRegistry.SetName("parameters");
	bInInstall = false;
	Containers.m_bUseLocking = true;
	m_iEncoding = ENCODING_NONE;
	m_bProtectContainers = false;
	m_sVariableIdent = "#";
}

NvService::~NvService()
{
	// Clear service information container
	m_Info.Clear();
	

	Stop();
}



void NvService::CleanSrvName(string& Name)
{
	MakeUpper(Name);
	CleanString(Name, true);
	string::size_type index;
	// remove backslashes
	while((index = Name.find('\\')) != string::npos)
		Name.erase(index, 1);
	// remove slashes
	while((index = Name.find('/')) != string::npos)
		Name.erase(index, 1);
	// remove double quotes
	while((index = Name.find('\"')) != string::npos)
		Name.erase(index, 1);
	// remove quotes
	while((index = Name.find('\'')) != string::npos)
		Name.erase(index, 1);
}

// Called from inside so it's not locked
bool NvService::CreateServiceDirectories()
{
	if(!m_sName.length())
		return false;
	
	
	// Create service dir for nirva home
	string SrvDirName = GetMain()->GetNirvaDir();
	SrvDirName += "Services";
	if(!CreateDir(SrvDirName.c_str()))	// In fact, the directory should exist
		return false;
	AddDirSuffix(SrvDirName);
	SrvDirName += m_sName;
	if(!CreateDir(SrvDirName.c_str()))
		return false;
	AddDirSuffix(SrvDirName);

	// Create service dir for log home
	string SrvHomeLogDirName = GetMain()->GetLogHomeDir();
	SrvHomeLogDirName += "Services";
	if(!CreateDir(SrvHomeLogDirName.c_str()))	// In fact, the directory should exist
		return false;
	AddDirSuffix(SrvHomeLogDirName);
	SrvHomeLogDirName += m_sName;
	if(!CreateDir(SrvHomeLogDirName.c_str()))
		return false;
	AddDirSuffix(SrvHomeLogDirName);

	// Create service dir for registry home
	string SrvHomeRegDirName = GetMain()->GetRegistryHomeDir();
	SrvHomeRegDirName += "Services";
	if(!CreateDir(SrvHomeRegDirName.c_str()))	// In fact, the directory should exist
		return false;
	AddDirSuffix(SrvHomeRegDirName);
	SrvHomeRegDirName += m_sName;
	if(!CreateDir(SrvHomeRegDirName.c_str()))
		return false;
	AddDirSuffix(SrvHomeRegDirName);
	
	// Create registry dir
	m_sRegistryDir = SrvHomeRegDirName;
	m_sRegistryDir += "Registry";
	if(!CreateDir(m_sRegistryDir.c_str()))
	{
		m_sRegistryDir = "";
		return false;
	}
	AddDirSuffix(m_sRegistryDir);

	// Create doc dir
	m_sDocDir = SrvDirName;
	m_sDocDir += "Docs";
	if(!CreateDir(m_sDocDir.c_str()))
	{
		m_sDocDir = "";
		return false;
	}
	else
	{
		// Also create the Html subdirectory
		string HtmlDir = m_sDocDir;
		AddDirSuffix(HtmlDir);
		HtmlDir += "Html";
		if(!CreateDir(HtmlDir.c_str()))
			return false;
	}
	AddDirSuffix(m_sDocDir);

	// And Html subdir

	// Create file dir
	m_sFileDir = SrvDirName;
	m_sFileDir += "Files";
	if(!CreateDir(m_sFileDir.c_str()))
	{
		m_sFileDir = "";
		return false;
	}
	else
	{
		// Also create the Config subdirectory
		string ConfigDir = m_sFileDir;
		AddDirSuffix(ConfigDir);
		ConfigDir += "Config";
		if(!CreateDir(ConfigDir.c_str()))
			return false;

		// And the Test and Test/Sets subdiretories
		string TestDir = m_sFileDir;
		AddDirSuffix(TestDir);
		TestDir += "Test";
		CreateDir(TestDir.c_str());
		AddDirSuffix(TestDir);
		TestDir += "Sets";
		CreateDir(TestDir.c_str());
	}
	AddDirSuffix(m_sFileDir);

	// Create log dir
	m_sLogDir = SrvHomeLogDirName;
	m_sLogDir += "Logs";
	if(!CreateDir(m_sLogDir.c_str()))
	{
		m_sLogDir = "";
		return false;
	}
	AddDirSuffix(m_sLogDir);

	// Create proc dir
	m_sProcDir = SrvDirName;
	m_sProcDir += "Procs";
	if(!CreateDir(m_sProcDir.c_str()))
	{
		m_sProcDir = "";
		return false;
	}
	else
	{
		// Also create the Config subdirectory
		string ConfigDir = m_sProcDir;
		AddDirSuffix(ConfigDir);
		ConfigDir += "Config";
		if(!CreateDir(ConfigDir.c_str()))
			return false;
	}
	AddDirSuffix(m_sProcDir);
	

	// Create bin dir
	m_sBinDir = SrvDirName;
	m_sBinDir += "Bin";
	if(!CreateDir(m_sBinDir.c_str()))
	{
		m_sBinDir = "";
		return false;
	}
	AddDirSuffix(m_sBinDir);

	// Create wroot dir
	m_sWrootDir = SrvDirName;
	m_sWrootDir += "Wroot";
	if(!CreateDir(m_sWrootDir.c_str()))
	{
		m_sWrootDir = "";
		return false;
	}
	else
	{
		// Also create the Config subdirectory
		string ConfigDir = m_sWrootDir;
		AddDirSuffix(ConfigDir);
		ConfigDir += "Config";
		if(!CreateDir(ConfigDir.c_str()))
			return false;
	}
	AddDirSuffix(m_sWrootDir);


	return true;
}


bool NvService::SetParameters(const char *Name, const char *Desc, const char *Path, bool Multithread, bool LoadOnStart, int Language)
{
	if(Name == NULL)
		return false;

	NV_LOCKOBJECT(this);
	string NameHere = Name;
	CleanString(NameHere);
	MakeUpper(NameHere);
	if(!NameHere.length())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	
	m_sName = NameHere;
	string DescHere = "";
	if(Desc != NULL)
		DescHere = Desc;
	CleanString(DescHere);
	m_sDesc = DescHere;
	m_bMultithread = Multithread;
	m_bLoadOnStart = LoadOnStart;
	m_iLanguage = Language;
	if(m_iLanguage == LANGUAGE_JAVA)
		hLib = GetMain()->JavaInterface.hLibJava;
	if(m_iLanguage == LANGUAGE_DOTNET)
		hLib = GetMain()->DotnetInterface.hLibDotnet;
	
	m_sPath = "";
	if(Path != NULL)
		m_sPath = Path;
	CleanString(m_sPath);

	SrvRegistry.SetName("parameters");

	if(!CreateServiceDirectories())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	string ServiceNameLower = m_sName;
	MakeLower(ServiceNameLower);

	// We must load now the description file if it exists
	LoadInfo();

	// Check if the service is supposed to be a renderer
	m_bIsRenderer = false;
	NvObjIndStringList *IndStringList = (NvObjIndStringList *)(m_Info.GetObject("SETTINGS"));
	if(IndStringList != NULL)
	{
		string sIsRenderer = IndStringList->GetValue("RENDERER");
		CleanString(sIsRenderer, true);
		MakeUpper(sIsRenderer);
		if(sIsRenderer == "YES")
			m_bIsRenderer = true;
	}

	if(!m_sPath.length())
	{
		// If path not given, we try to get it from the dsc file
		if(IndStringList != NULL)
		{
			m_sPath = IndStringList->GetValue("PATH");
			CleanString(m_sPath, true);
		}
/*		if(m_sPath.length())
		{
			string SrvDirName = GetMain()->GetNirvaDir();
			SrvDirName += "Services";
			AddDirSuffix(SrvDirName);
			SrvDirName += m_sName;
			AddDirSuffix(SrvDirName);
			SrvDirName += m_sPath;
			m_sPath = SrvDirName;
		}*/
	}

	if(m_iLanguage == LANGUAGE_JAVA)
	{
		if(!m_sPath.length())
		{
			m_sPath = "Bin";
			AddDirSuffix(m_sPath);
			m_sPath += ServiceNameLower;
		}
	}
	else if(m_iLanguage == LANGUAGE_DOTNET)
	{
		if(!m_sPath.length())
		{
			m_sPath = "Bin";
			AddDirSuffix(m_sPath);
			m_sPath += ServiceNameLower;
			m_sPath += ".dll";
		}
	}
	else
	{
		if(!m_sPath.length())
		{
			m_sPath = "Bin";
			AddDirSuffix(m_sPath);
			m_sPath += ServiceNameLower;
#ifdef WIN32
			m_sPath += ".dll";
#endif
#ifdef LINUX
			m_sPath += ".so";
#endif
#ifdef HPUX
			m_sPath += ".sl";
#endif
#ifdef HPUXI
			m_sPath += ".so";
#endif
#ifdef SOLARIS
			m_sPath += ".so";
#endif
#ifdef AIX
			m_sPath += ".a";
#endif
		}

		// We check that the file exists
		string PathHere = GetRealPath();
		if(!GetNoCaseRealFileName(PathHere))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		if(!FileExist(PathHere.c_str()))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}


	// We set the service root directory
	SrvRegistry.SetRootDirectory(m_sRegistryDir.c_str());

	NV_UNLOCKOBJECT(this);

	return true;
}




// Load service information from dsc file
bool NvService::LoadInfo()
{
	// The description file should be in the File directory
	string ServiceNameLower = m_sName;
	MakeLower(ServiceNameLower);
	string DescFileName = m_sFileDir;
	DescFileName += "service.dsc";

	// If the service.dsc doesn't exist, use the real service name (compatibility with old versions)
	if(!GetNoCaseRealFileName(DescFileName) || !FileExist(DescFileName.c_str()))
	{
		DescFileName = m_sFileDir;
		DescFileName += ServiceNameLower;
		DescFileName += ".dsc";
	}

	if(GetNoCaseRealFileName(DescFileName))
		if(FileExist(DescFileName.c_str()))
			if(!m_Info.Load(DescFileName.c_str()))
				return false;

	m_iEncoding = ENCODING_NONE;
	NvObjIndStringList *IndStringList = (NvObjIndStringList *)(m_Info.GetObject("INFO"));
	if(IndStringList != NULL)
	{
		if(!m_sDesc.length())
		{
			if(IndStringList->KeyExist("DESCRIPTION"))
				m_sDesc = IndStringList->GetValue("DESCRIPTION");
			else if(IndStringList->KeyExist("Description"))
				m_sDesc = IndStringList->GetValue("Description");
			else if(IndStringList->KeyExist("description"))
				m_sDesc = IndStringList->GetValue("description");
		}
		string sEncoding = IndStringList->GetValue("ENCODING");
		CleanString(sEncoding);
		MakeUpper(sEncoding);
		if(sEncoding == "UTF-8")
			m_iEncoding = ENCODING_UTF8;
		if(sEncoding == "ISO-8859-1")
			m_iEncoding = ENCODING_LATIN1;
	}
	m_sVariableIdent = "#";	// default value
	IndStringList = (NvObjIndStringList *)(m_Info.GetObject("SETTINGS"));
	if(IndStringList != NULL)
	{
		if(IndStringList->KeyExist("Variable_Ident"))
			m_sVariableIdent = IndStringList->GetValue("Variable_Ident");
	}
	return true;
}


bool NvService::CreateSourceSkeleton(bool WithSessions, int Language, const char *SessionId)
{
	NV_LOCKOBJECT(this);
	

	if(!CreateServiceDirectories())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	// Create source dir
	string ServiceSourceDir = GetMain()->GetNirvaDir();
	ServiceSourceDir += "Services";
	AddDirSuffix(ServiceSourceDir);
	ServiceSourceDir += m_sName;
	AddDirSuffix(ServiceSourceDir);
	ServiceSourceDir += "Source";
	if(!CreateDir(ServiceSourceDir.c_str()))
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	AddDirSuffix(ServiceSourceDir);

	NvObjFile SourceFile;
	SourceFile.SetPersistTime();
	NvObjFile DestinationFile;
	DestinationFile.SetPersistTime();
	
	
	string FileDir;
	string FileName;
	string ServiceNameLower = m_sName;
	MakeLower(ServiceNameLower);
	
	if(Language == LANGUAGE_CPP)
	{
		// We must first copy the nvsext.h file
		FileDir = GetMain()->GetNirvaDir();
		FileDir += "Sdk";
		AddDirSuffix(FileDir);
		FileDir += "Service";
		AddDirSuffix(FileDir);
		FileName = FileDir;
		FileName += "nvsext.h";
		SourceFile.SetFileName(FileName);
		
		FileName = ServiceSourceDir;
		FileName += "nvsext.h";
		DestinationFile.SetFileName(FileName);

		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			// Just used here to remove /r characters if they are here
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}
	}

	FileDir = GetMain()->GetNirvaDir();
	FileDir += "Files";
	AddDirSuffix(FileDir);
	FileDir += "Service";
	AddDirSuffix(FileDir);
	FileName = FileDir;

	// Then we copy the license.txt file
	FileDir += "Skeleton";
	AddDirSuffix(FileDir);

	FileName = FileDir;
	FileName += "license.txt";
	SourceFile.SetFileName(FileName);


	FileName = ServiceSourceDir;
	FileName += "license.txt";
	DestinationFile.SetFileName(FileName);

	// .def file
	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	if(Language == LANGUAGE_JAVA)
	{
		FileName = FileDir;
		FileName += "service.java";
		SourceFile.SetFileName(FileName);


		FileName = ServiceSourceDir;
		FileName += ServiceNameLower;
		FileName += ".java";
		DestinationFile.SetFileName(FileName);

		// .def file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}
	}
	if(Language == LANGUAGE_CPP)
	{
		// Then we copy the cpp and header files

		FileName = FileDir;
		FileName += "service.def";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += ServiceNameLower;
		FileName += ".def";
		DestinationFile.SetFileName(FileName);

		// .def file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "service.h";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += ServiceNameLower;
		FileName += ".h";
		DestinationFile.SetFileName(FileName);

		// .h file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "service.cpp";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += ServiceNameLower;
	#ifdef AIX
		FileName += ".C";
	#else
		FileName += ".cpp";
	#endif
		DestinationFile.SetFileName(FileName);

		// .cpp file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}

		FileName = FileDir;
		FileName += "nvthread.h";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += "nvthread.h";
		DestinationFile.SetFileName(FileName);

		// .h file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "nvthread.cpp";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
	#ifdef AIX
		FileName += "nvthread.C";
	#else
		FileName += "nvthread.cpp";
	#endif
		DestinationFile.SetFileName(FileName);

		// .cpp file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "service.dsw";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += ServiceNameLower;
		FileName += ".dsw";
		DestinationFile.SetFileName(FileName);

		// .dsw file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "service.dsp";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += ServiceNameLower;
		FileName += ".dsp";
		DestinationFile.SetFileName(FileName);

		// .dsp file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}
	}	// End C++ specific

	if(Language == LANGUAGE_DOTNET)
	{
		// Then we copy the cs file

		FileName = FileDir;
		FileName += "dotnet.cs";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += ServiceNameLower;
		FileName += ".cs";
		DestinationFile.SetFileName(FileName);

		// .cs file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "dotnetcompile.bat";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += "compile.bat";
		DestinationFile.SetFileName(FileName);

		// compile.bat file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}

#ifdef WIN32
		// For windows we also generate the visual C# 2008 project
		string GuidProject = CreateGuid();
		string GuidSolution = CreateGuid();
		string GuidTypeLib = CreateGuid();

		FileName = FileDir;
		FileName += "dotnet.sln";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += ServiceNameLower;
		FileName += ".sln";
		DestinationFile.SetFileName(FileName);

		// .sln file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId, GuidProject.c_str(), GuidSolution.c_str(), GuidTypeLib.c_str()))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}

		FileName = FileDir;
		FileName += "dotnet.csproj";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += ServiceNameLower;
		FileName += ".csproj";
		DestinationFile.SetFileName(FileName);

		// .csproj file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId, GuidProject.c_str(), GuidSolution.c_str(), GuidTypeLib.c_str()))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "dotnetassemblyinfo.cs";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += "Properties";
		if(!CreateDir(FileName.c_str()))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		AddDirSuffix(FileName);
		FileName += "assemblyinfo.cs";
		DestinationFile.SetFileName(FileName);

		// assemblyinfo.cs file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId, GuidProject.c_str(), GuidSolution.c_str(), GuidTypeLib.c_str()))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}
#endif
	}	// End dotnet specific

	// Description file
	FileName = FileDir;
	FileName += "service.dsc";
	SourceFile.SetFileName(FileName);

	FileName = m_sFileDir;
	FileName += "service.dsc";
	DestinationFile.SetFileName(FileName);

	// .dsc file
	if(!DestinationFile.Exist())
	{
		// Check if there is no old service name description file
		string oldDescFileName = m_sFileDir;
		oldDescFileName += ServiceNameLower;
		oldDescFileName += ".dsc";
		if(!FileExist(oldDescFileName.c_str()))
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}
	}

	// Default doc file
	FileName = FileDir;
	FileName += "doc.htm";
	SourceFile.SetFileName(FileName);

	FileName = m_sDocDir;
	FileName += "Html";
	AddDirSuffix(FileName);
	FileName += ServiceNameLower;
	FileName += ".htm";
	DestinationFile.SetFileName(FileName);

	// default.htm file
	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	// Config procedure file
	FileName = FileDir;
	FileName += "config.nvp";
	SourceFile.SetFileName(FileName);

	FileName = m_sProcDir;
	FileName += "Config";
	AddDirSuffix(FileName);
	FileName += "config.nvp";
	DestinationFile.SetFileName(FileName);

	// .nvp file
	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	// Config xsl file
	FileName = FileDir;
	FileName += "config.xsl";
	SourceFile.SetFileName(FileName);

	FileName = m_sFileDir;
	FileName += "Config";
	AddDirSuffix(FileName);
	FileName += "config.xsl";
	DestinationFile.SetFileName(FileName);

	// .xsl file
	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}


	// Install procedure file
	FileName = FileDir;
	FileName += "install.nvp";
	SourceFile.SetFileName(FileName);

	FileName = m_sProcDir;
	FileName += "install.nvp";
	DestinationFile.SetFileName(FileName);

	// .nvp file
	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	// package file
	FileName = FileDir;
	FileName += "package.lst";
	SourceFile.SetFileName(FileName);

	FileName = m_sFileDir;
	FileName += "package.lst";
	DestinationFile.SetFileName(FileName);

	// .lst file
	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	if(Language == LANGUAGE_CPP)
	{
		FileName = FileDir;
		FileName += "makefile.aix";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += "makefile.aix";
		DestinationFile.SetFileName(FileName);

		// makefile.aix file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "makefile.linux";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += "makefile.linux";
		DestinationFile.SetFileName(FileName);

		// makefile.linux file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "makefile.hpux";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += "makefile.hpux";
		DestinationFile.SetFileName(FileName);

		// makefile.hpux file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}


		FileName = FileDir;
		FileName += "makefile.hpuxi";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += "makefile.hpuxi";
		DestinationFile.SetFileName(FileName);

		// makefile.hpuxi file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}

		FileName = FileDir;
		FileName += "makefile.solaris";
		SourceFile.SetFileName(FileName);

		FileName = ServiceSourceDir;
		FileName += "makefile.solaris";
		DestinationFile.SetFileName(FileName);

		// makefile.solaris file
		if(!DestinationFile.Exist())
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
			if(!SetSkeletonFile(&DestinationFile, WithSessions, Language, SessionId))
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}
	}


	NV_UNLOCKOBJECT(this);

	return true;
}

bool NvService::Package(const char *DescriptionFile, const char *PackageFile, string& Info)
{
	NV_LOCKOBJECT(this);

	if(!CreateServiceDirectories())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	string SrvDirName = GetMain()->GetNirvaDir();
	SrvDirName += "Services";
	AddDirSuffix(SrvDirName);
	SrvDirName += m_sName;
	AddDirSuffix(SrvDirName);

	string DescriptionFileInParam = "";
	if(DescriptionFile != NULL)
		DescriptionFileInParam = DescriptionFile;
	CleanString(DescriptionFileInParam);

	string DescriptionFileHere = SrvDirName;
	DescriptionFileHere += "Files";
	AddDirSuffix(DescriptionFileHere);
	if(DescriptionFileInParam.length())
		DescriptionFileHere += DescriptionFileInParam;
	else
		DescriptionFileHere += "package.lst";

	NvPackage Package;
	bool RetVal = Package.Check(DescriptionFileHere.c_str(), SrvDirName.c_str());
	if(RetVal)
		RetVal = Package.Create(DescriptionFileHere.c_str(), SrvDirName.c_str(), PackageFile);
	if(!RetVal)
		Info = Package.Info;

	NV_UNLOCKOBJECT(this);

	return RetVal;
}

bool NvService::Install(const char *PackageFile, string& Info)
{
	NV_LOCKOBJECT(this);

	if(!CreateServiceDirectories())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	string SrvDirName = GetMain()->GetNirvaDir();
	SrvDirName += "Services";
	AddDirSuffix(SrvDirName);
	SrvDirName += m_sName;
	AddDirSuffix(SrvDirName);

	NvPackage Package;
	bool RetVal = Package.Install(PackageFile, SrvDirName.c_str(), GetMain()->GetDebug());
	if(!RetVal)
		Info = Package.Info;

	NV_UNLOCKOBJECT(this);

	return RetVal;
}

// used for automatic source code generation
bool NvService::SetSkeletonFile(NvObjFile *File, bool WithSessions, int Language, const char *SessionId, const char *GuidProject, const char *GuidSolution, const char *GuidTypelib)
{

	NvSession *Session = GetMain()->Sessions.GetSession(SessionId);
	if(Session == NULL)
		return false;

	if(Session->Application == NULL)
	{
		GetMain()->Sessions.FreeSession(SessionId);
		return false;
	}
	
	string AppWorkDir = Session->Application->GetWorkDir();

	if(!File->Open("r+b"))
	{
		GetMain()->Sessions.FreeSession(SessionId);
		File->Remove();
		return false;
	}

	NvObjFile ProviFile;
	ProviFile.SetAutoFileName(AppWorkDir.c_str());
	if(!ProviFile.Open("w+b"))
	{
		GetMain()->Sessions.FreeSession(SessionId);
		File->Close();
		File->Remove();
		return false;
	}
	
	string Line;
	string::size_type index;
	bool SessionSpecific = false;
	bool NoSessionSpecific = false;
	bool WriteLine = false;

	string ServiceNameUpper = m_sName;
	MakeUpper(ServiceNameUpper);
	
	// remove spaces
	while((index = ServiceNameUpper.find(' ')) != string::npos)
		ServiceNameUpper.erase(index, 1);

	string ServiceNameLower = m_sName;
	MakeLower(ServiceNameLower);
	// remove spaces
	while((index = ServiceNameLower.find(' ')) != string::npos)
		ServiceNameLower.erase(index, 1);


	// Create service identifier
	char Buffer[128];
	sprintf(Buffer, "%d%d%d", myRand(), myRand(), myRand());
	string sIdentifier = Buffer;
	sIdentifier += m_sName;
	sIdentifier += Session->m_sIdent;
	sIdentifier = MD5_string(sIdentifier);
	MakeUpper(sIdentifier);
	string sIdentifierForCheck = MD5_string(sIdentifier);
	MakeUpper(sIdentifierForCheck);

	// Service language
	string sLanguage = "C++";
	if(Language == LANGUAGE_JAVA)
		sLanguage = "JAVA";
	if(Language == LANGUAGE_DOTNET)
		sLanguage = "DOTNET";

	GetMain()->Sessions.FreeSession(SessionId);

	while(File->ReadLine(Line))
	{
		if(Line.find("startsessionspecific") !=  string::npos)
		{
			SessionSpecific = true;
			continue;
		}
		if(Line.find("endsessionspecific") !=  string::npos)
		{
			SessionSpecific = false;
			continue;
		}
		if(Line.find("startnosessionspecific") !=  string::npos)
		{
			NoSessionSpecific = true;
			continue;
		}
		if(Line.find("endnosessionspecific") !=  string::npos)
		{
			NoSessionSpecific = false;
			continue;
		}

		if(WithSessions)
		{
			if(NoSessionSpecific)
				WriteLine = false;
			else
				WriteLine = true;
		}
		else
		{
			if(SessionSpecific)
				WriteLine = false;
			else
				WriteLine = true;
		}

		if(!WriteLine)
			continue;

		while((index = Line.find("upperservicename")) != string::npos)
			Line.replace(index, 16, ServiceNameUpper);
		while((index = Line.find("lowerservicename")) != string::npos)
			Line.replace(index, 16, ServiceNameLower);
		while((index = Line.find("nirvaserviceidcheck")) != string::npos)
			Line.replace(index, 19, sIdentifierForCheck);
		while((index = Line.find("nirvaserviceid")) != string::npos)
			Line.replace(index, 14, sIdentifier);
		while((index = Line.find("servicelanguage")) != string::npos)
			Line.replace(index, 15, sLanguage);
		if(GuidProject != NULL)
			while((index = Line.find("csserviceprojectguid")) != string::npos)
				Line.replace(index, 20, GuidProject);
		if(GuidSolution != NULL)
			while((index = Line.find("csservicesolutionguid")) != string::npos)
				Line.replace(index, 21, GuidSolution);
		if(GuidTypelib != NULL)
			while((index = Line.find("csservicetypelibguid")) != string::npos)
				Line.replace(index, 20, GuidTypelib);

#ifdef WIN32
		Line += "\r\n";
#else
		Line += "\n";
#endif
		
		if(ProviFile.Write(Line.c_str(), (int)(Line.length())) != (int)(Line.length()))
		{
			ProviFile.Close();
			File->Close();
			File->Remove();
			return false;
		}
	}

	ProviFile.Close();
	File->Close();
	
	if(!File->CopyFrom(&ProviFile))
	{
		File->Remove();
		return false;
	}
		
	return true;
}



bool NvService::IsInitialized()
{
	bool RetVal = true;
	NV_LOCKOBJECT(this);
	RetVal = Initialized;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

bool NvService::IsStarted()
{
	bool RetVal = true;
	NV_LOCKOBJECT(this);
	RetVal = Started;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

bool NvService::Init()
{
	NV_LOCKOBJECT(this);
	if(Initialized)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}

	if(!CreateServiceDirectories())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	// We must load now the description file if it exists
	LoadInfo();

	// We set the service root directory
	SrvRegistry.SetRootDirectory(m_sRegistryDir.c_str());

	TestSets.m_sDirPath = m_sFileDir;
	TestSets.m_sDirPath += "Test";
	AddDirSuffix(TestSets.m_sDirPath);
	TestSets.m_sDirPath += "Sets";
	TestSets.Load();


	if(m_iLanguage == LANGUAGE_CPP)
	{
		// We check that the file exists
		string PathHere = GetRealPath();
		if(!GetNoCaseRealFileName(PathHere))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		if(!FileExist(PathHere.c_str()))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	Initialized = true;

	NV_UNLOCKOBJECT(this);

	return true;
}


bool NvService::Start()
{
	if(!Init())
		return false;

	NV_LOCKOBJECT(this);
	if(Started)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}


	if(!IsEnable())
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}

	string LogMess = "Starting service ";
	LogMess += m_sName;
	GetMain()->Logs.WriteFlush("System", LogMess.c_str(), "", NV_LOG_INFO, 1, "", "system");
	if(GetMain()->b_ConsoleMode)
		printf("%s\n", LogMess.c_str());

	// We check that the file exists
	if(m_iLanguage == LANGUAGE_CPP)
	{
		string PathHere = GetRealPath();
		if(!GetNoCaseRealFileName(PathHere))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		if(!FileExist(PathHere.c_str()))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		// Load the service here
		hLib = MyLoadLibrary(PathHere.c_str());
	}
	if(hLib == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	LibCommand = (NVSCOMMANDENTRY)MyGetLibFunction(hLib, "NvsCommand");

	if(LibCommand == NULL)
	{
		if(m_iLanguage == LANGUAGE_CPP)
		{
			MyFreeLibrary(hLib);
			hLib = NULL;
		}
		NV_UNLOCKOBJECT(this);
		return false;
	}

	// We have started the service

	// Get the optional MvsRender function if it exists (for rendering services)
	LibRender = (NVSRENDERENTRY)MyGetLibFunction(hLib, "NvsRender");

	// we then send the NV_INIT_SERVICE command
	NvSession *Session = NVNEW NvSession;
	Session->Init();
	Session->Login(NULL, NULL, "", SESSION_TYPE_SYSTEM);	// Login with default application and super internal user 
	Session->m_bAdministrator = true;
	NvCommand RealCommand;
	Session->OriginalCommand = &RealCommand;
	RealCommand.m_pSession = Session;
	MyNvServiceCommand ServiceCommand;
	if(m_iLanguage == LANGUAGE_JAVA)
		ServiceCommand.i_Encoding = ENCODING_UTF8;
	else if(m_iLanguage == LANGUAGE_DOTNET)
		ServiceCommand.i_Encoding = ENCODING_UTF8;
	else
		ServiceCommand.i_Encoding = m_iEncoding;
	ServiceCommand.m_pCommand = &RealCommand;
	RealCommand.m_iCommandSource = NV_SOURCE_SERVER;
	RealCommand.m_sService = m_sName;
	RealCommand.m_sClass = "SYSTEM";
	RealCommand.m_sCommand = "NV_INIT_SERVICE";

	if(m_iLanguage == LANGUAGE_JAVA)
	{
//		RealCommand.m_sCommand += "#&NV&#";
//		RealCommand.m_sCommand += m_sName;
		RealCommand.Parameters.SetValue("NV_JAVA_SERVICE_NAME", m_sName);
		RealCommand.Parameters.SetValue("NV_JAVA_SERVICE_PATH", GetRealPath());
		SetJavaLogLevel(GetMain()->Logs.GetLevel("JAVA"));
	}
	if(m_iLanguage == LANGUAGE_DOTNET)
	{
//		RealCommand.m_sCommand += "#&NV&#";
//		RealCommand.m_sCommand += m_sName;
		RealCommand.Parameters.SetValue("NV_DOTNET_SERVICE_NAME", m_sName);
		RealCommand.Parameters.SetValue("NV_DOTNET_SERVICE_PATH", GetRealPath());
	}
	
	int Success = 0;
	try
	{
		Success = (LibCommand)((void *)(&ServiceCommand));
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		GetMain()->Logs.Write("Exceptions", "Exception in init service", m_sName.c_str(), NV_LOG_ERROR);
		Success = 0;
	}

	// Session is not need any more
	delete Session;

	if(!Success)
	{
		// The INIT_SERVICE failed
		if(m_iLanguage == LANGUAGE_CPP)
		{
			MyFreeLibrary(hLib);
			hLib = NULL;
		}
		NV_UNLOCKOBJECT(this);
		return false;
	}

	// Do init code here
	Started = true;
	LastSessionId = "";
	NV_UNLOCKOBJECT(this);
	return true;
}

bool NvService::Stop(int WaitTime)
{
	NV_LOCKOBJECT(this);
	if(Started)
	{

		string LogMess = "Stopping service ";
		LogMess += m_sName;
		GetMain()->Logs.WriteFlush("System", LogMess.c_str(), "", NV_LOG_INFO, 1, "", "system");
		if(GetMain()->b_ConsoleMode)
			printf("%s\n", LogMess.c_str());
		if(WaitTime < 0)
			WaitTime = 0;

		if(WaitTime > 0)
		{
			// Wait first for the service to be free
			// But a maximum of WaitTime
			// We check every 200 ms

			int i;
			bool InUseHere = true;
			int NumLoop = (1000 * WaitTime)/200;
			for(i=0; i<NumLoop; i++)
			{
				if(!GetInUse())
				{
					InUseHere = false;
					break;
				}
				SLEEP(200);
			}

			if(InUseHere)
			{
				NV_UNLOCKOBJECT(this);
				return false;
			}
		}

		// we then send the NV_EXIT_SERVICE command
		NvSession *Session = NVNEW NvSession;
		Session->Init();
		Session->Login(NULL, NULL, "", SESSION_TYPE_SYSTEM);	// Login with default application and super internal user 
		Session->m_bAdministrator = true;
		NvCommand RealCommand;
		Session->OriginalCommand = &RealCommand;
		RealCommand.m_pSession = Session;
		MyNvServiceCommand ServiceCommand;
		if(m_iLanguage == LANGUAGE_JAVA)
			ServiceCommand.i_Encoding = ENCODING_UTF8;
		else if(m_iLanguage == LANGUAGE_DOTNET)
			ServiceCommand.i_Encoding = ENCODING_UTF8;
		else
			ServiceCommand.i_Encoding = m_iEncoding;
		ServiceCommand.m_pCommand = &RealCommand;
		RealCommand.m_iCommandSource = NV_SOURCE_SERVER;
		RealCommand.m_sService = m_sName;
		RealCommand.m_sClass = "SYSTEM";
		RealCommand.m_sCommand = "NV_EXIT_SERVICE";
		if(m_iLanguage == LANGUAGE_JAVA)
		{
//			RealCommand.m_sCommand += "#&NV&#";
//			RealCommand.m_sCommand += m_sName;
			RealCommand.Parameters.SetValue("NV_JAVA_SERVICE_NAME", m_sName);
			SetJavaLogLevel(GetMain()->Logs.GetLevel("JAVA"));
		}
		if(m_iLanguage == LANGUAGE_DOTNET)
		{
//			RealCommand.m_sCommand += "#&NV&#";
//			RealCommand.m_sCommand += m_sName;
			RealCommand.Parameters.SetValue("NV_DOTNET_SERVICE_NAME", m_sName);
		}
		try
		{
			(LibCommand)((void *)(&ServiceCommand));	// the command is not necessary implemented into the service
		}
		catch(...)
		{
			GetMain()->IncNumExceptions();
			GetMain()->Logs.Write("Exceptions", "Exception in exit service", m_sName.c_str(), NV_LOG_ERROR);
		}

		NV_LOCKLOCK(&ProtectInUse);
		InUse = 0;
		NV_UNLOCKLOCK(&ProtectInUse);

		delete Session; // Not needed any more

		// Unload the service here
		if(m_iLanguage == LANGUAGE_CPP)
		{
			MyFreeLibrary(hLib);
			hLib = NULL;
		}
		LibCommand = NULL;
		LibRender = NULL;
	}
	
	// Do cleanup code here
	LastSessionId = "";
	Started = false;

	SrvRegistry.Clear();		// Removes the service registry containers from memory
	Containers.Clear();			// Clear service containers
	TestSets.Clear();			// Clear the test set list

	// Free the dotnet AppDomain
	GetMain()->DotnetInterface.UnloadDomain(DOMAIN_TYPE_SERVICE, m_sName.c_str());
	// Free the java environment
	GetMain()->JavaInterface.UnloadEnv(JAVAENV_TYPE_SERVICE, m_sName.c_str());

	Initialized = false;

	NV_UNLOCKOBJECT(this);

	return true;
}


// Check if this service is required for some applications
// Not locked because its called from the stop function
bool NvService::CheckRequiredForApplications()
{
	return GetMain()->Applications.IsRequiredService(m_sName);
}


bool NvService::IncInuse()
{
	NV_LOCKLOCK(&ProtectInUse);
//	if(!m_bEnable)
//	{
//		NV_UNLOCKLOCK(&ProtectInUse);
//		return false;
//	}
	InUse++;
	NV_UNLOCKLOCK(&ProtectInUse);

	return true;
}

void NvService::DecInUse()
{
	NV_LOCKLOCK(&ProtectInUse);
	if(InUse)
		InUse--;
	NV_UNLOCKLOCK(&ProtectInUse);
}

int NvService::GetInUse()
{
	int RetVal = 0;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = InUse;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


bool NvService::Enable(int StartIt)
{
	if(!Init())
		return false;
	NV_LOCKLOCK(&ProtectInUse);
	m_bEnable = true;
	NV_UNLOCKLOCK(&ProtectInUse);

	bool RetVal = true;
	bool bStartIt = false;
	if(StartIt == START_MODE_START)
		bStartIt = true;
	if(StartIt == START_MODE_FLAG)
		bStartIt = m_bLoadOnStart;

	if(bStartIt)
		RetVal = Start();

	return RetVal;
}


// Disable the service and unload it from memory
// SessionId is the sesion id of the session launching the Disable service command
bool NvService::Disable(const char *SessionId, int WaitTime)
{
	SetDisableFlag();
	SLEEP(100);
	
	NV_LOCKOBJECT(this);
	if(!Started)
	{
		Initialized = false;
		NV_UNLOCKOBJECT(this);
		return Stop(WaitTime);	// The stop function is doing the necessary cleaning
	}

	// Now we should send NV_EXIT_SESSION command to the service
	// for all connected sessions

	string SessionIdHere = "";
	if(SessionId != NULL)
		SessionIdHere = SessionId;
	CleanString(SessionIdHere);
	MakeUpper(SessionIdHere);
	map<string,string>::iterator pos;
	for(pos = Sessions.m_mList.begin(); pos != Sessions.m_mList.end(); ++pos)
	{
		string SessionIdInList = pos->first;
		NvSession *Session = GetMain()->Sessions.GetSession(SessionIdInList.c_str());
		if(Session == NULL)
			continue;
		// We lock the session if it's not the calling session (because the calling session is ever locked)
		if(Session->m_sIdent != SessionIdHere)
			NV_LOCKOBJECT(Session);;

		NvCommand RealCommand;
		RealCommand.m_pSession = Session;
		MyNvServiceCommand ServiceCommand;
		if(m_iLanguage == LANGUAGE_JAVA)
			ServiceCommand.i_Encoding = ENCODING_UTF8;
		else if(m_iLanguage == LANGUAGE_DOTNET)
			ServiceCommand.i_Encoding = ENCODING_UTF8;
		else
			ServiceCommand.i_Encoding = m_iEncoding;
		ServiceCommand.m_pCommand = &RealCommand;
		RealCommand.m_iCommandSource = NV_SOURCE_SERVER;
		RealCommand.m_sService = m_sName;
		RealCommand.m_sClass = "SYSTEM";
		RealCommand.m_sCommand = "NV_EXIT_SESSION";
		if(m_iLanguage == LANGUAGE_JAVA)
		{
//			RealCommand.m_sCommand += "#&NV&#";
//			RealCommand.m_sCommand += m_sName;
			RealCommand.Parameters.SetValue("NV_JAVA_SERVICE_NAME", m_sName);
			SetJavaLogLevel(GetMain()->Logs.GetLevel("JAVA"));
		}
		if(m_iLanguage == LANGUAGE_DOTNET)
		{
//			RealCommand.m_sCommand += "#&NV&#";
//			RealCommand.m_sCommand += m_sName;
			RealCommand.Parameters.SetValue("NV_DOTNET_SERVICE_NAME", m_sName);
		}
		try
		{
			(LibCommand)((void *)(&ServiceCommand));	// the command is not necessary implemented into the service
		}
		catch(...)
		{
			GetMain()->IncNumExceptions();
			GetMain()->Logs.Write("Exceptions", "Exception in init service session", m_sName.c_str(), NV_LOG_ERROR);
		}

		// Close all opened user registries for this session and this service
		map<string,int>::iterator pos2;
		vector<string> RegistriesToErase;
		int i;
		for(pos2 = Session->m_oRegistries.begin(); pos2 != Session->m_oRegistries.end(); ++pos2)
		{
			string RegistryName = pos2->first;
			string RegistryService = "";
			string::size_type index = RegistryName.find("::::");
			if(index != string::npos)
			{
				RegistryService = RegistryName.substr(0, index);
				RegistryName = RegistryName.substr(index+4);
			}
			CleanString(RegistryService);
			MakeUpper(RegistryService);
			int NumInstances =  pos2->second;

			string ServiceNameHere = m_sName;
			MakeUpper(ServiceNameHere);
			if(RegistryService != RegistryService)
				continue;

			for(i=0;i<NumInstances;i++)
				Registries.CloseRegistry(RegistryName);
			
			RegistriesToErase.push_back(pos2->first);
		}
		for(i=0;i<(int)(RegistriesToErase.size());i++)
			Session->m_oRegistries.erase(RegistriesToErase[i]);

		// We do not remove the service from the list of services connected
		// by the session (Session->m_oServices)
		// In this way, the session will generate an error message if trying to connect again
		// to the service
		// If we don't do that, the next time the session will try to connect to the re-enabled
		// service, the service creates a new session for it and the session doesn't see that
		// the service has been unloaded and then reloaded. This is not good.
		// Except for the calling session

		if(Session->m_sIdent == SessionIdHere)
		{
			NV_LOCKLOCK(&(Session->ProtectInUse));	// The access to m_oServices is protected by the InUse Lock for the session reporting
			int i;
			for(i=0; i<Session->m_oServices.GetSize(); i++)
			{
				if(Session->m_oServices.GetValue(i) == m_sName)
				{
					Session->m_oServices.Remove(i);
					break;
				}
			}
			NV_UNLOCKLOCK(&(Session->ProtectInUse));
		}
		else
			NV_UNLOCKOBJECT(Session);

		GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
	}
	Sessions.RemoveAll();

	// Clear all user registries for this service
	Registries.Clear();

	// Disable the container protection (can be enabled only via a service command)
	SetProtectContainers(false);
	NV_UNLOCKOBJECT(this);

	return Stop(WaitTime);	// We stop the service (if started) at disable time
}

bool NvService::IsEnable()
{
	bool RetVal = true;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_bEnable;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

bool NvService::IsRenderer()
{
	bool RetVal = false;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_bIsRenderer;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

void NvService::SetDisableFlag()
{
	NV_LOCKLOCK(&ProtectInUse);
	m_bEnable = false;
	NV_UNLOCKLOCK(&ProtectInUse);
}

void NvService::SetProtectContainers(bool ProtectContainers)
{
	NV_LOCKLOCK(&ProtectInUse);
	m_bProtectContainers = ProtectContainers;
	NV_UNLOCKLOCK(&ProtectInUse);
}

bool NvService::GetProtectContainers()
{
	bool RetVal = true;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_bProtectContainers;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvService::GetName()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sName;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}

string NvService::GetDesc()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sDesc;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}


string NvService::GetPath()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sPath;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}

// Get the complete path name following what we have in m_sPath (relate or complete)
string NvService::GetRealPath(bool bLockIt)
{
	if(bLockIt)
		NV_LOCKOBJECT(this);
	string ServicePath;
	if(bLockIt)
		ServicePath = GetPath();
	else
		ServicePath = m_sPath;
	string RetVal = ServicePath;
	string SrvDirName = GetMain()->GetNirvaDir();
	SrvDirName += "Services";
	AddDirSuffix(SrvDirName);
	SrvDirName += m_sName;
	AddDirSuffix(SrvDirName);
	string SrvDirNameHere = SrvDirName;
	
	// replace backslach with slash
	string::size_type index;
	while((index = SrvDirNameHere.find('\\')) != string::npos)
		SrvDirNameHere.replace(index, 1, 1, '/');
	
	string PathHere = RetVal;
	while((index = PathHere.find('\\')) != string::npos)
		PathHere.replace(index, 1, 1, '/');
#ifdef WIN32
	MakeUpper(SrvDirNameHere);
	MakeUpper(PathHere);
#endif

	if(PathHere.find(SrvDirNameHere) != 0)
	{
		// The path is relative
		RetVal = SrvDirName;
		RetVal += ServicePath;
	}

	if(bLockIt)
		NV_UNLOCKOBJECT(this);

	return RetVal;
}

string NvService::GetRegistryDir()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sRegistryDir;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}


string NvService::GetDocDir()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sDocDir;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}


string NvService::GetFileDir()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sFileDir;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}


string NvService::GetLogDir()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sLogDir;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}


string NvService::GetProcDir()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sProcDir;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}


string NvService::GetBinDir()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sBinDir;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}


string NvService::GetWrootDir()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sWrootDir;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}



string NvService::GetVariableIdent()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sVariableIdent;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

bool NvService::GetSessions(NvObjStringList *ConnectedSessions)
{
	ConnectedSessions->RemoveAll();
	NV_LOCKOBJECT(this);
	map<string,string>::iterator pos;
	for(pos = Sessions.m_mList.begin(); pos != Sessions.m_mList.end(); ++pos)
	{
		string SessionIdInList = pos->first;
		ConnectedSessions->Add(SessionIdInList);
	}
	NV_UNLOCKOBJECT(this);

	return true;
}


// Each command sent to a service goes here
// The commands are serialized (locked) if the service is not multithread
// The session init and exit commands are also serialized in any case. This avoid for
// the service builder to implement a locking mechanism for special init and exit sessions
// commands. The other system commands sent to the service are also serialized with the same lock
// object.

bool NvService::Command(MyNvServiceCommand *pCommand)
{
	NvCommand *RealCommand = (NvCommand *)(pCommand->m_pCommand);
	NvSession *Session = (NvSession *)(RealCommand->m_pSession);
	if(!Init())
	{
		// We must set an error code here
		RealCommand->Error.SetError("SYSTEM", "SERVICE", NV_SERVICE_CANNOT_INIT, GetName());
		return false;
	}
	if(!IsEnable())
	{
		// We must set an error code here
		RealCommand->Error.SetError("SYSTEM", "SERVICE", NV_SERVICE_STOPPED, GetName());
		return false;
	}
	if(!Start())	// Start the service if not started
	{
		// We must set an error code here
		RealCommand->Error.SetError("SYSTEM", "SERVICE", NV_SERVICE_CANNOT_INIT, GetRealPath(true).c_str());
		return false;
	}

	bool bSessionInit = false;
	bool bSessionExit = false;
	if(RealCommand->m_iCommandSource == NV_SOURCE_SERVER)
	{
		if((RealCommand->m_sClass == "SYSTEM") && (RealCommand->m_sCommand == "NV_INIT_SESSION"))
			bSessionInit = true;
		if((RealCommand->m_sClass == "SYSTEM") && (RealCommand->m_sCommand == "NV_EXIT_SESSION"))
			bSessionExit = true;
	}

	// if the command is the first command for this session
	// we must send an INIT_SESSION first
	string ActualSessionId = Session->m_sIdent.c_str();
	NV_LOCKOBJECT(this);
	if(!Sessions.KeyExist(ActualSessionId))
	{
		// we add the session to the list
		Sessions.SetValue(ActualSessionId, "");
		// This is the first session call

		// Check if the NV_INIT_SESSION command is not sent by the session directly
		if(!bSessionInit)
		{
			// So we send it ourself
			NvCommand SpecialRealCommand;
			SpecialRealCommand.m_pSession = Session;
			MyNvServiceCommand ServiceCommand;
			if(m_iLanguage == LANGUAGE_JAVA)
				ServiceCommand.i_Encoding = ENCODING_UTF8;
			else if(m_iLanguage == LANGUAGE_DOTNET)
				ServiceCommand.i_Encoding = ENCODING_UTF8;
			else
				ServiceCommand.i_Encoding = m_iEncoding;
			ServiceCommand.m_pCommand = &SpecialRealCommand;
			SpecialRealCommand.m_iCommandSource = NV_SOURCE_SERVER;
			SpecialRealCommand.m_sService = m_sName;
			SpecialRealCommand.m_sClass = "SYSTEM";
			SpecialRealCommand.m_sCommand = "NV_INIT_SESSION";
			if(m_iLanguage == LANGUAGE_JAVA)
			{
//				SpecialRealCommand.m_sCommand += "#&NV&#";
//				SpecialRealCommand.m_sCommand += m_sName;
				SpecialRealCommand.Parameters.SetValue("NV_JAVA_SERVICE_NAME", m_sName);
				SetJavaLogLevel(GetMain()->Logs.GetLevel("JAVA"));
			}
			if(m_iLanguage == LANGUAGE_DOTNET)
			{
//				SpecialRealCommand.m_sCommand += "#&NV&#";
//				SpecialRealCommand.m_sCommand += m_sName;
				SpecialRealCommand.Parameters.SetValue("NV_DOTNET_SERVICE_NAME", m_sName);
			}
			int Success = 1;
			try
			{
				Success = (LibCommand)((void *)(&ServiceCommand));	// the command is not necessary implemented into the service
			}
			catch(...)
			{
				GetMain()->IncNumExceptions();
				GetMain()->Logs.Write("Exceptions", "Exception in init service session", m_sName.c_str(), NV_LOG_ERROR);
				Success = false;
			}
			if(!Success)
			{
				// We must set an error code here
				NV_UNLOCKOBJECT(this);
				RealCommand->Error.SetError("SYSTEM", "SERVICE", NV_SERVICE_CANNOT_INIT_SESSION, GetName().c_str());
				return false;
			}
			// We set the service session ID as value of the Sessions indexed string list
			Sessions.SetValue(ActualSessionId, ServiceCommand.m_sServiceSessionId);
		}
	}
	NV_UNLOCKOBJECT(this);

	bool bLocked = false;

	// If not multithread, we lock the service commands.
	// In order to keep the possibility for a service to call itself, we
	// implemented a mechanism that locks only when the session calling the service
	// is different than the previous one
	if(!m_bMultithread)
	{
		if(ActualSessionId != LastSessionId)
		{
			NV_LOCKOBJECT(this);
			bLocked = true;
			LastSessionId = ActualSessionId;
		}
	}
	if((bSessionInit || bSessionExit) && (!bLocked))
	{
		NV_LOCKOBJECT(this);	// We must lock for the exit and init session commands
		bLocked = true;
	}

	// We send back the service session ID to the command
	if(!bSessionInit)
		pCommand->m_sServiceSessionId = Sessions.GetValue(ActualSessionId);

	
	bool RetVal = true;
	int Success = 1;
	
//	string OriCommand = "";
	if(m_iLanguage == LANGUAGE_JAVA)
	{
//		OriCommand = RealCommand->m_sCommand;
//		RealCommand->m_sCommand += "#&NV&#";
//		RealCommand->m_sCommand += m_sName;
		RealCommand->Parameters.SetValue("NV_JAVA_SERVICE_NAME", m_sName);
		SetJavaLogLevel(GetMain()->Logs.GetLevel("JAVA"));
	}
	if(m_iLanguage == LANGUAGE_DOTNET)
	{
//		OriCommand = RealCommand->m_sCommand;
//		RealCommand->m_sCommand += "#&NV&#";
//		RealCommand->m_sCommand += m_sName;
		RealCommand->Parameters.SetValue("NV_DOTNET_SERVICE_NAME", m_sName);
	}

	try
	{
		Success = (LibCommand)((void *)pCommand);
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		GetMain()->Logs.Write("Exceptions", "Exception in service command", m_sName.c_str(), NV_LOG_ERROR);
//		if(m_iLanguage == LANGUAGE_JAVA)
//			RealCommand->m_sCommand = OriCommand;
//		if(m_iLanguage == LANGUAGE_DOTNET)
//			RealCommand->m_sCommand = OriCommand;
		RetVal = false;
//		if(bLocked)
//			NV_UNLOCKOBJECT(this);
//		throw;		// Rethrow exception
	}

//	if(m_iLanguage == LANGUAGE_JAVA)
//		RealCommand->m_sCommand = OriCommand;
//	if(m_iLanguage == LANGUAGE_DOTNET)
//		RealCommand->m_sCommand = OriCommand;
	if(!Success)
		RetVal = false;


	// If init session, We set the service session ID as value of the Sessions indexed string list
	if(RetVal && bSessionInit)
	{
		Sessions.SetValue(ActualSessionId, pCommand->m_sServiceSessionId);
	}


	if(bSessionExit)
	{
		// We remove the session from the list if we received a SYSTEM NV_EXIT_SESSION command
		// We are necessary ever locked for that
		Sessions.Remove(ActualSessionId);
	}

	if(bLocked)
		NV_UNLOCKOBJECT(this);

	return RetVal;
}



bool NvService::Render(MyNvRendererInt *RendererInterface, NvSession *Session)
{

	if(!Init())
	{
		RendererInterface->m_sReason = "Renderer service not initialized";
		return false;
	}
	if(!IsEnable())
	{
		RendererInterface->m_sReason = "Renderer service not started";
		return false;
	}
	if(!Start())	// Start the service if not started
	{
		RendererInterface->m_sReason = "Cannot start renderer service";
		return false;
	}
	if(LibRender == NULL)
	{
		RendererInterface->m_sReason = "The service is not a renderer service";
		return false;
	}

	// if the command is the first command for this session
	// we must send an INIT_SESSION first
	string ActualSessionId = Session->m_sIdent;
	NV_LOCKOBJECT(this);
	if(!Sessions.KeyExist(ActualSessionId))
	{
		// we add the session to the list
		Sessions.SetValue(ActualSessionId, "");
		// This is the first session call

		NvCommand SpecialRealCommand;
		SpecialRealCommand.m_pSession = Session;
		MyNvServiceCommand ServiceCommand;
		if(m_iLanguage == LANGUAGE_JAVA)
			ServiceCommand.i_Encoding = ENCODING_UTF8;
		else if(m_iLanguage == LANGUAGE_DOTNET)
			ServiceCommand.i_Encoding = ENCODING_UTF8;
		else
			ServiceCommand.i_Encoding = m_iEncoding;
		ServiceCommand.m_pCommand = &SpecialRealCommand;
		SpecialRealCommand.m_iCommandSource = NV_SOURCE_SERVER;
		SpecialRealCommand.m_sService = m_sName;
		SpecialRealCommand.m_sClass = "SYSTEM";
		SpecialRealCommand.m_sCommand = "NV_INIT_SESSION";
		if(m_iLanguage == LANGUAGE_JAVA)
		{
//				SpecialRealCommand.m_sCommand += "#&NV&#";
//				SpecialRealCommand.m_sCommand += m_sName;
			SpecialRealCommand.Parameters.SetValue("NV_JAVA_SERVICE_NAME", m_sName);
			SetJavaLogLevel(GetMain()->Logs.GetLevel("JAVA"));
		}
		if(m_iLanguage == LANGUAGE_DOTNET)
		{
//				SpecialRealCommand.m_sCommand += "#&NV&#";
//				SpecialRealCommand.m_sCommand += m_sName;
			SpecialRealCommand.Parameters.SetValue("NV_DOTNET_SERVICE_NAME", m_sName);
		}
		int Success = 1;
		try
		{
			Success = (LibCommand)((void *)(&ServiceCommand));	// the command is not necessary implemented into the service
		}
		catch(...)
		{
			GetMain()->IncNumExceptions();
			GetMain()->Logs.Write("Exceptions", "Exception in renderer service", m_sName.c_str(), NV_LOG_ERROR);
			Success = false;
		}
		if(!Success)
		{
			// We must set an error code here
			NV_UNLOCKOBJECT(this);
			RendererInterface->m_sReason = "Cannot intialize renderer service session";
			return false;
		}
		// We set the service session ID as value of the Sessions indexed string list
		Sessions.SetValue(ActualSessionId, ServiceCommand.m_sServiceSessionId);
	}
	NV_UNLOCKOBJECT(this);

	bool bLocked = false;

	// If not multithread, we lock the service commands.
	// In order to keep the possibility for a service to call itself, we
	// implemented a mechanism that locks only when the session calling the service
	// is different than the previous one
	if(!m_bMultithread)
	{
		if(ActualSessionId != LastSessionId)
		{
			NV_LOCKOBJECT(this);
			bLocked = true;
			LastSessionId = ActualSessionId;
		}
	}
	
	bool RetVal = true;
	int Success = 1;
	
	try
	{
		Success = (LibRender)((void *)RendererInterface, m_sName.c_str());
	}
	catch(...)
	{
		GetMain()->IncNumExceptions();
		GetMain()->Logs.Write("Exceptions", "Exception in renderer service", m_sName.c_str(), NV_LOG_ERROR);
		RetVal = false;
	}

	if(!Success)
		RetVal = false;


	if(bLocked)
		NV_UNLOCKOBJECT(this);

	return RetVal;
}


/////////////////////////////////////////////////////////////////////////////
// MyNvRendererInt

MyNvRendererInt::MyNvRendererInt()
{
	m_pInputData = NULL;
	m_uInputDataSize = 0;
	m_sPage = "";
	m_pOutputData = NULL;
	m_uOutputDataSize = 0;
	m_sReason = "";
	m_sContentType = "";
	m_sHttpRet = "";
	m_sSessionId = "";
	m_sAppName = "";
	i_Encoding = ENCODING_NONE;
}

MyNvRendererInt::~MyNvRendererInt()
{
	if(m_pInputData != NULL)
		free(m_pInputData);
	m_pInputData = NULL;
	m_uInputDataSize = 0;
	if(m_pOutputData != NULL)
		free(m_pOutputData);
	m_pOutputData = NULL;
	m_uOutputDataSize = 0;
}

const char* MyNvRendererInt::GetInputData()
{
	return (const char*)m_pInputData;
}


unsigned long MyNvRendererInt::GetInputDataSize()
{
	return m_uInputDataSize;
}


const char* MyNvRendererInt::GetInputEncoding()
{
	return m_sInputEncoding.c_str();
}


const char* MyNvRendererInt::GetPageName()
{
	return m_sPage.c_str();
}

// Alloc, realloc or free output buffer
char* MyNvRendererInt::AllocOutputBuffer(unsigned long Size)
{
	if(Size == 0)
	{
		// We free the output buffer
		if(m_pOutputData != NULL)
			free(m_pOutputData);
		m_pOutputData = NULL;
		m_uOutputDataSize = 0;
		return NULL;
	}

	if(m_pOutputData == NULL)
	{
		// Alloc
		m_pOutputData = (char*)malloc(Size);
		if(m_pOutputData != NULL)
			m_uOutputDataSize = Size;
	}
	else
	{
		// Realloc
		char *NewOutputBuffer = (char *)realloc(m_pOutputData, Size);
		if(NewOutputBuffer == NULL)
		{
			// Realloc failed
			free(m_pOutputData);
			m_pOutputData = NULL;
			m_uOutputDataSize = 0;
			return NULL;
		}
		m_pOutputData = NewOutputBuffer;
		m_uOutputDataSize = Size;
	}

	return m_pOutputData;
}

char* MyNvRendererInt::GetOutputBuffer()
{
	return m_pOutputData;
}

unsigned long MyNvRendererInt::GetOutputBufferSize()
{
	return m_uOutputDataSize;
}

const char* MyNvRendererInt::GetSessionId()
{
	return m_sSessionId.c_str();
}

const char* MyNvRendererInt::GetAppName()
{
	return m_sAppName.c_str();
}

void MyNvRendererInt::SetReason(const char* Reason)
{
	if(Reason != NULL)
		m_sReason = Reason;
}

void MyNvRendererInt::SetContentType(const char* ContentType)
{
	if(ContentType != NULL)
		m_sContentType = ContentType;
}

void MyNvRendererInt::SetHttpRet(const char* HttpRet)
{
	if(HttpRet != NULL)
		m_sHttpRet = HttpRet;
}

void* MyNvRendererInt::GetExtraData(const char* DataName)
{
	return NULL;	// For futur extension
}

void MyNvRendererInt::FreeInputData()
{
	if(m_pInputData != NULL)
		free(m_pInputData);
	m_pInputData = NULL;
	m_uInputDataSize = 0;
}



