// ***************************************************************************************
// File: webs.h
// Description: declarations of the Nirva web service classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifndef _WEBS_H
#define _WEBS_H

// Classes declared in this file
class NvWebsContent;					// The web service content
class NvWebs;							// A single web service
class NvWebsList;						// A web service list


#include "task.h"
#include "container.h"
#include <map>

#define WEB_SERVICE_VERSION			"1.00"

/////////////////////////////////////////////////////////////////////////////
// NvWebsContent class
class NvWebsContent
{
public:
	NvWebs *WebService;									// Pointer to the owner web service
	std::map<std::string, NvContainer*> m_mMessages;	// Messages
	NvObjTable m_tOperations;							// Operations
	std::string m_sVersion;								// Version of the web service
	std::string m_sHeaders;								// Tags for soapheader
		
public:
	NvWebsContent();
	virtual ~NvWebsContent();

	void Clear();
	void CopyFrom(NvWebsContent *Source);
	bool Load();
	bool Save();
	
	bool OperationList(NvObjTable *OpList, std::string AppName);
	bool OperationEnable(const char *Operation, bool Enable);
	bool OperationCreate(const char *Operation, const char *Description, const char *Procedure, const char *InMessage, const char *OutMessage);
	bool OperationRemove(const char *Operation);
	bool OperationSetParam(const char *Operation, const char *Description, const char *Procedure, const char *InMessage, const char *OutMessage);
	bool OperationGetParam(const char *Operation, std::string &Description, std::string &Procedure, std::string &InMessage, std::string &OutMessage);
	bool OperationExist(const char *Operation);
	bool OperationSoapIn(const char *Operation, NvObjBinary *Output, std::string AppName, std::string Host);
	std::string MessageGetRealName(const char *Message);
	std::string MessageGetDescription(const char *Message);
	NvContainer* MessageGetInformationContainer(const char *Message);
	NvContainer* MessageGetDataContainer(const char *Message);
	bool MessageList(NvObjTable *MsgList);
	bool MessageCreate(const char *Message, const char *Description);
	bool MessageRemove(const char *Message);
	bool MessageSetParam(const char *Message, const char *Description);
	bool MessageSetContent(const char *Message, NvContainer* ContentCont);
	bool MessageGetContent(const char *Message, NvContainer* ContentCont);
	bool SetHeaders(const char *Headers);
	std::string GetHeaders();
};


/////////////////////////////////////////////////////////////////////////////
// NvWebs class
class NvWebs : public NvObject
{
private:
	int InUse;									// In use counter
	NvMutex ProtectInUse;						// protect locking of web service for updating the in use counter
	bool Initialized;
	bool CreateWebsDirectories();				// Creates web service directories if they don't exist

public:
	std::string m_sName;
	std::string m_sDesc;
	std::string m_sNsPrefix;					// Namespace prefix (server name)
	bool m_bEnable;								// Web service enabled or not
	bool m_bHttpsOnly;							// Execution in https only
	bool m_bUseDefaultPath;						// Use default path (appname/websname)
	std::string m_sPath;						// Path
	std::string m_sBinDir;						// Webs directory
	std::string m_sFileDir;						// File directory
	std::string m_sProcDir;						// Procedure directory
	std::string m_sDocDir;						// Procedure directory
	NvDescriptionContainer m_Info;				// Web service information
	bool bInInstall;							// In installation process
	NvSession *m_sSessionEditMode;				// Session that owns the edit mode
	NvWebsContent Content;						// Web service content
	NvWebsContent CopyContent;					// Web service copy content (used in edit mode)
	NvWebsContent *CurrentContent;				// Current web service work content (content or copy)
		
public:
	NvWebs();
	virtual ~NvWebs();

	bool Init();
	void Cleanup();

	void IncInuse();
	void DecInUse();
	int GetInUse();

	bool Enable();
	bool Disable();
	bool IsEnable();

	bool EnterEditMode(NvSession *Session);
	bool ExitEditMode(NvSession *Session);
	NvSession* GetSessionEditMode();
	bool EditUpdate(NvSession *Session, std::string& Info);

	bool SetParameters(const char *Name, const char *Desc, const char *NsPrefix, bool HttpsOnly, bool UseDefaultPath, const char *Path);
	bool Package(const char *DescriptionFile, const char *PackageFile, std::string& Info);
	bool Install(const char *PackageFile, std::string& Info);

	std::string GetName();
	std::string GetDesc();
	std::string GetNsPrefix();
	std::string GetPath();
	std::string GetBinDir();
	std::string GetDocDir();
	std::string GetFileDir();
	std::string GetProcDir();
	bool GetHttpsOnly();
	bool GetUseDefaultPath();
	bool SetWebsFile(NvObjFile *File);
	bool LoadInfo();
	static void CleanWebsName(std::string& Name);
	static void ParseHostName(std::string Host, std::vector<std::string>& Protocols, std::vector<std::string>& Hosts); 

	bool SetWsdl(std::string AppName, std::string Host, NvObjBinary *Result);
	bool SetWsdlMessage(const char *Message, NvObjBinary *Result, bool Input = false, void *XmlPtr = NULL);
	bool SetWsdlFault(NvObjBinary *Result);
	bool SetWsdlCommandHeader(NvObjBinary *Result, void *XmlPtr = NULL);

	bool OperationList(NvObjTable *OpList, NvSession *Session);
	bool OperationEnable(const char *Operation, bool Enable);
	bool OperationCreate(const char *Operation, const char *Description, const char *Procedure, const char *InMessage, const char *OutMessage);
	bool OperationRemove(const char *Operation);
	bool OperationSetParam(const char *Operation, const char *Description, const char *Procedure, const char *InMessage, const char *OutMessage);
	bool OperationSoapIn(const char *Operation, NvObjBinary *Output, NvSession *Session, std::string AppName, std::string Host);
	bool MessageList(NvObjTable *MsgList, NvSession *Session);
	bool MessageCreate(const char *Message, const char *Description);
	bool MessageRemove(const char *Message);
	bool MessageSetParam(const char *Message, const char *Description);
	bool MessageSetContent(const char *Message, NvContainer* ContentCont);
	bool MessageGetContent(const char *Message, NvContainer* ContentCont, NvSession *Session);
	bool SetHeaders(const char *Headers);
	std::string GetHeaders(NvSession *Session);
};



/////////////////////////////////////////////////////////////////////////////
// NvWebsList class
class NvWebsList : public NvObject
{
private:
	NvWebs* InternalGetWebs(const char *Name);

public:
	std::map<std::string, NvWebs*> m_mWebservices;
	std::map<std::string, NvSession*> m_mWebservicesInEdit;
		
public:
	NvWebsList();
	virtual ~NvWebsList();

	// Removes all Web services from the list
	void Clear();
	
	// used to get an existing Web service
	// The function increments the in use counter of the corresponding Web service if found and if LockIt is set to true
	NvWebs* GetWebservice(const char *Name, bool LockIt = true, bool GetIfNotStarted = false);

	// Load the Web service list from registry
	void Load();

	// Saves the Web service list to the system registry
	bool Save();

	// Creates a new Web service with the given name if it doesn't exist
	// Fails if the Web service ever exists and is running
	bool CreateWebservice(const char *Name, const char *Desc, const char *NsPrefix, bool HttpsOnly, bool UseDefaultPath, const char *Path, bool FromLoad = false);


	// used to free an existing Web service
	// The function decrements the in use counter of the corresponding Web service
	void FreeWebservice(const char *Name);
	
	// Remove the given Web service from the list
	bool RemoveWebservice(const char *Name);

	// Enable a Web service (default)
	bool EnableWebservice(const char *Name, NvSession *Session, bool FromLoad = false);

	// Disable a Web service
	bool DisableWebservice(const char *Nameint, int WaitTime = 60);

	// Checks if a Web service exists in the list
	bool WebserviceExist(const char *Name);

	// Checks if a Web service is enabled or not
	bool IsWebserviceEnable(const char *Name);

	// Get Web service info
	bool GetInfo(const char *Name, NvObjTable *WebsInfo, NvSession *Session, std::string Host);

	// Get Web service specific info
	bool GetInfoEx(const char *Name, NvObjIndStringList *WebsInfo);

	// Package Web service
	bool Package(const char *Name, const char *DescriptionFile, const char *PackageFile, std::string& Info);

	// Install Web service
	bool Install(const char *Name, const char *PackageFile, std::string& Info);

	// Set edit mode
	bool SetEditMode(const char *Name, NvSession *Session, bool Edit = true);

	// Test if the actual session is in edit mode
	bool IsInEdit(const char *Name, NvSession *Session);

	// Test if the session can set the edit mode
	bool CanEdit(const char *Name, NvSession *Session);

	// Update the web service content
	bool EditUpdate(const char *Name, NvSession *Session, std::string& Info);

	// Free the edited web services owned by the session
	bool FreeEdit(NvSession *Session);

	// List the operations for a web service
	bool OperationList(const char *Name, NvObjTable *OpList, NvSession *Session);

	// Enable or disable an operation
	bool OperationEnable(const char *Name, const char *Operation, bool Enable);

	// Create a new operation
	bool OperationCreate(const char *Name, const char *Operation, const char *Description, const char *Procedure, const char *InMessage, const char *OutMessage);

	// Remove an operation
	bool OperationRemove(const char *Name, const char *Operation);

	// Set operation parameters
	bool OperationSetParam(const char *Name, const char *Operation, const char *Description, const char *Procedure, const char *InMessage, const char *OutMessage);

	// Set an example of SOAP input message
	bool OperationSoapIn(const char *Name, const char *Operation, NvObjBinary *Output, NvSession *Session, std::string AppName, std::string Host);

	// List the messages for a web service
	bool MessageList(const char *Name, NvObjTable *MsgList, NvSession *Session);

	// Create a new message
	bool MessageCreate(const char *Name, const char *Message, const char *Description);

	// Remove a message
	bool MessageRemove(const char *Name, const char *Message);

	// Set message parameters
	bool MessageSetParam(const char *Name, const char *Message, const char *Description);

	// Set message content
	bool MessageSetContent(const char *Name, const char *Message, NvContainer* ContentCont);

	// Get message content
	bool MessageGetContent(const char *Name, const char *Message, NvContainer* ContentCont, NvSession *Session);

	// Set headers
	bool SetHeaders(const char *Name, const char *Headers);
	
	// Get headers
	std::string GetHeaders(const char *Name, NvSession *Session);

	// Set the WSDL content for a web service
	bool SetWsdl(const char *Name, std::string AppName, std::string Host, NvObjBinary *Result);

	// Get all web service permissions
	bool GetPermissions(NvObjTable *WebsPermissions);
};



#endif //_WEBS_H
