// ***************************************************************************************
// File: xml.h
// Description: declarations of the Nirva xml classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifndef _XML_H
#define _XML_H


// Classes declared in this file
class NvXml;


#include "task.h"
#include "container.h"
#include "session.h"
#include <libxml/xmlmemory.h>
#include <libxml/debugXML.h>
#include <libxml/HTMLtree.h>
#include <libxml/xmlIO.h>
#include <libxml/DOCBparser.h>
#include <libxml/xinclude.h>
#include <libxml/catalog.h>
#include <libxml/xmlschemas.h>
#include <libxml/xmlschemastypes.h>
#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>
#include <libxslt/extensions.h>
#include <libexslt/exslt.h>


// Initilialisation and cleanup of the XML library
void InitXml();
void CleanupXml();

/////////////////////////////////////////////////////////////////////////////
// NvXml class
class NvXml
{
private:
	bool Initialized;
	xmlDocPtr Doc;
	xmlCharEncodingHandlerPtr EncodingHandler;

public:
	
	bool bStrict;
		
	NvXml();
	virtual ~NvXml();

	void FreeDoc();
	bool XsltTransform(const char *SourceFile, const char *DestinationFile, const char *XsltFile, NvObjIndStringList *Parameters = NULL, bool SkipDtd = false);
	bool XsltTransform(char **Data, unsigned long *DataSize, const char *XsltFile, NvObjIndStringList *Parameters = NULL, bool SkipDtd = false);
	bool XsltTransform(std::string &Data, std::string &XsltFile, NvObjIndStringList *Parameters = NULL, bool SkipDtd = false);
	bool SetInputData(NvContainer *Container, NvCommand *CommandToExecute, std::string& NameSpace, int& XmlCoding);
	bool SetInputData(NvContainer *Container, const char *Data, unsigned long DataSize, NvCommand *CommandToExecute, std::string& NameSpace, int& XmlCoding);
	bool PopulateContainer(NvContainer *Container, NvCommand *CommandToExecute, xmlNodePtr Node, int& XmlCoding);
	bool PopulateObject(NvContainer *Container, NvCommand *CommandToExecute, xmlNodePtr Node, const char* ObjName);
	void EncodeInputString(std::string& StringToEncode);
	std::string EncodeOutputString(std::string StringToEncode, bool bUnicode = false);
	bool GetCommandParam(NvObjIndStringList* Parameters, const char* Data, int DataSize, std::string& NameSpace);
	bool GetSoapHeaderCommandParam(NvObjIndStringList* Parameters, const char* Data, int DataSize);
	bool ValidateSchema(const char* XmlDocFile, const char* XmlSchemaFile, std::string& ResultData);
	bool ValidateSchema(char *Data, unsigned long DataSize, const char *XmlSchemaFile, std::string& ResultData);
	static void ValidityError(void *ctx, const char *msg, ...);
	bool ValidateDtd(const char* XmlDocFile, const char* XmlDtdFile, std::string& ResultData);
	bool ValidateDtd(char *Data, unsigned long DataSize, const char *XmlDtdFile, std::string& ResultData);

	static void EnableCache(bool Enable);
	static void EmptyCache();
};


/////////////////////////////////////////////////////////////////////////////
// NvXslStyleSheet class
class NvXslStyleSheet : public NvObject
{
public:

	std::string m_sXsltFile;
	std::string m_sLastModifiedDate;
	xsltStylesheetPtr m_xStyleSheet;
	int m_iInUse;

	NvXslStyleSheet();
	virtual ~NvXslStyleSheet();

	// Removes all stylesheets from the list
	void Clear();
};


/////////////////////////////////////////////////////////////////////////////
// NvXslCache class
class NvXslCache : public NvObject
{
public:
	bool m_bUseCache;
	NvMutex ProtectUseCache;
		
	std::map<std::string,NvXslStyleSheet*> m_mStyleSheets;	// Test list

	NvXslCache();
	virtual ~NvXslCache();

	void Enable(bool bEnable);
	bool IsEnable();
	
	// Removes all stylesheets from the list
	void Clear();

	// Empty the cache
	void Empty();

	// Get a specific style sheet and add it if not exist
	xsltStylesheetPtr GetStyleSheet(const char *XsltFile, bool& InCache);
	
	// Free a stylesheet
	void FreeStyleSheet(const char *XsltFile);
};


#endif //_XML_H
